import datetime
from abc import abstractmethod
from typing import List, Union, Dict

import gitlab
from loguru import logger

from robota_core import gitlab_tools, config_readers
from robota_core.github_tools import GithubServer
from robota_core.merge_request import MergeRequest, MergeRequestCache


class RemoteProvider:
    """A remote provider is a cloud provider that a git repository can be synchronised to.
    Remote providers have some features that a basic git Repository does not including merge
    requests and teams.
    """
    def __init__(self):
        self._stored_merge_requests: List[MergeRequestCache] = []

    def get_merge_requests(self, start: datetime.datetime = datetime.datetime.fromtimestamp(1, datetime.timezone.utc),
                           end: datetime.datetime = datetime.datetime.now(tz=datetime.timezone.utc)) -> List[MergeRequest]:
        cached_merge_requests = self._get_cached_merge_requests(start, end)
        if cached_merge_requests:
            return cached_merge_requests.merge_requests

        new_merge_requests = self._fetch_merge_requests(start, end)
        cache = MergeRequestCache(start, end, new_merge_requests)
        self._stored_merge_requests.append(cache)
        return new_merge_requests

    @abstractmethod
    def _fetch_merge_requests(self, start: datetime.datetime,
                              end: datetime.datetime) -> List[MergeRequest]:
        raise NotImplementedError("Not implemented in base class")

    def _get_cached_merge_requests(self, start: datetime.datetime,
                                   end: datetime.datetime) -> Union[MergeRequestCache, None]:
        """Check whether merge requests with the specified start and end date are already stored."""
        for cache in self._stored_merge_requests:
            if cache.start == start and cache.end == end:
                return cache
        else:
            return None

    @abstractmethod
    def get_members(self) -> Dict[str, str]:
        """Get a dictionary of names and corresponding usernames of members of this repository."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_wiki_pages(self) -> Dict[str, str]:
        """Get a dictionary of page slugs to page contents from the wiki of this repository"""
        raise NotImplementedError("Not implemented in base class.")


class GithubRemoteProvider(RemoteProvider):
    def __init__(self, provider_source: dict):
        super().__init__()
        server = GithubServer(provider_source)
        self.repo = server.open_github_repo(provider_source["project"])

    def _fetch_merge_requests(self, start: datetime.datetime,
                              end: datetime.datetime) -> List[MergeRequest]:
        all_pulls = self.repo.get_pulls()
        filtered_pulls = [pull for pull in all_pulls if start < pull.created_at < end]
        return [MergeRequest(pull, "github") for pull in filtered_pulls]

    def get_members(self) -> Dict[str, str]:
        """This method returns names and usernames of repo collaborators since github doesn't
        have the idea of members in the same way as gitlab."""
        members = self.repo.get_collaborators()
        member_names = {member.name: member.login for member in members}
        return member_names

    def get_wiki_pages(self) -> Dict[str, str]:
        raise NotImplementedError("Not implemented in base class.")


class GitlabRemoteProvider(RemoteProvider):
    def __init__(self, provider_source: dict):
        super().__init__()
        if "token" in provider_source:
            token = provider_source["token"]
        else:
            token = None
        server = gitlab_tools.GitlabServer(provider_source["url"], token)
        self.project = server.open_gitlab_project(provider_source["project"])

        super().__init__()

    def _fetch_merge_requests(self, start: datetime.datetime,
                              end: datetime.datetime) -> List[MergeRequest]:
        """Get merge requests within a time period"""
        merge_requests = self.project.mergerequests.list(created_after=start, created_before=end)
        return [MergeRequest(merge_request, "gitlab") for merge_request in merge_requests]

    def get_members(self) -> Dict[str, str]:
        members = self.project.members.list()
        member_names = {member.attributes['name']: member.attributes['username']
                        for member in members}
        return member_names

    def get_wiki_pages(self) -> Dict[str, Dict[str, str]]:
        """Get a dictionary of page slugs to page contents from the wiki of this repository.
        GitLab doesn't return the page contents as part of the list of wikis returned from the
        project.  These have to be requested individually given the slug."""
        wiki_pages_by_slug = {}
        pages = self.project.wikis.list()
        for page in pages:
            slug = page.slug
            title = page.title
            try:
                contents = self.project.wikis.get(slug).content
                if len(contents) > 1000:
                    contents = contents[:999]
                wiki_pages_by_slug[slug] = {'title': title,
                                            'content': contents,
                                            'url': self.project.web_url + '/-/wikis/' + slug}
            except gitlab.exceptions.GitlabGetError:
                logger.warning('Wiki page could not be returned: ' + slug)

        return wiki_pages_by_slug


def new_remote_provider(robota_config: dict) -> Union[RemoteProvider, None]:
    """Factory method for RemoteProvider."""
    provider_config = config_readers.get_data_source_info(robota_config, "remote_provider")
    if not provider_config:
        return None
    provider_type = provider_config["type"]

    logger.debug(f"Initialising {provider_type} remote provider.")

    if provider_type == "gitlab":
        return GitlabRemoteProvider(provider_config)
    elif provider_type == "github":
        return GithubRemoteProvider(provider_config)
    else:
        raise TypeError(f"Unknown remote provider type {provider_config['type']}.")



"""
Module that defines interactions with a Continuous integration server in order to get
build information.
"""
import json
from loguru import logger
from datetime import datetime, timezone
from enum import Enum
from typing import Union, List, Dict, TypeVar
from abc import ABC, abstractmethod

import jenkins
import requests

from robota_core.string_processing import string_to_datetime, get_link
from robota_core import config_readers


class Test:
    """A representation of the result of a Test.

    :ivar name: The name of the test.
    :ivar result: The result of the test, PASSED or FAILED.
    :ivar time: The time that the test ran.
    :ivar branch: The branch of commit the test was run upon. This is not populated on object
      creation."""
    def __init__(self, suite: dict, case: dict):
        self.name = f"{suite['name']}.{case['name']}"
        self.result = case["status"]
        self.time = string_to_datetime(suite["timestamp"], "%Y-%m-%dT%H:%M:%S")
        self.branch = None

    def __eq__(self, test_result: "Test"):
        if self.name == test_result.name:
            return True
        return False

    def __hash__(self):
        return hash(self.name)


class BuildResult(Enum):
    """Represents the result of a Jenkins build."""
    Success = 1
    Unstable = 2
    Failure = 3
    Aborted = 4
    Not_Built = 5
    Gitlab_Timeout = 6

    def __str__(self):
        return self.name


class Build:
    """A Build is the result of executing a CI job.

    :ivar number: The number of the build.
    :ivar result: The result of the build
    :ivar timestamp: The time at which the build started.
    :ivar commit_id: The ID of the git commit the build was run on.
    :ivar branch_name: The git branch of the commit the build was run on.
    :ivar link: A HTML string linking to the web-page that displays the build on Jenkins.
    :ivar instruction_coverage: A code coverage result from JaCoCo.
    """
    def __init__(self, jenkins_build):
        self.number: str = ""
        self.result: BuildResult = None
        self.timestamp: datetime = None
        self.commit_id: str = ""
        self.branch_name: str = ""
        self.link: str = ""
        self.instruction_coverage: dict = {}
        self.test_coverage_url: str = ""

        self.build_from_jenkins(jenkins_build)

    def build_from_jenkins(self, jenkins_build):
        """Create a Robota Build object from a Jenkins build object."""
        self.number = jenkins_build["number"]
        self.result = self._assign_build_result(jenkins_build["result"])
        self.timestamp = datetime.fromtimestamp(jenkins_build["timestamp"] / 1000, timezone.utc)
        self.link = get_link(jenkins_build["url"], self.result.name)
        self.test_coverage_url = f'{jenkins_build["url"]}jacoco/'
        for action in jenkins_build["actions"]:
            if "_class" in action:
                if action["_class"] == "hudson.plugins.git.util.BuildData":
                    self.commit_id = action["lastBuiltRevision"]["SHA1"]
                    self.branch_name = action["lastBuiltRevision"]["branch"][0]["name"]
                if action["_class"] == "hudson.plugins.jacoco.JacocoBuildAction":
                    if "instructionCoverage" in action:
                        self.instruction_coverage = action['instructionCoverage']
                if "FailureCauseBuildAction" in action["_class"]:
                    for cause in action["foundFailureCauses"]:
                        if cause["name"] == 'Connection time-out while accessing GitLab':
                            self.result = BuildResult.Gitlab_Timeout

    @staticmethod
    def _assign_build_result(build_result: str) -> BuildResult:
        """Convert the build result string from Jenkins into a BuildResult representation."""
        if build_result == "SUCCESS":
            return BuildResult.Success
        elif build_result == "UNSTABLE":
            return BuildResult.Unstable
        elif build_result == "FAILURE":
            return BuildResult.Failure
        elif build_result == "ABORTED":
            return BuildResult.Aborted
        elif build_result == "NOT_BUILT" or build_result is None:
            return BuildResult.Failure
        else:
            raise KeyError(f"Build result of type {build_result} not known.")


class Job:
    """A job is a series of CI checks. Each time a job is executed it stores
    the result in a build.
    """
    def __init__(self, job_data, project_root):
        self.name = None
        self.short_name = None
        self.url = None
        # Builds are ordered most recent first.
        self.last_build_number = None
        self.last_completed_build_number = None
        self.last_successful_build_number = None
        self._builds: List[Build] = []

        self.job_from_jenkins(job_data, project_root)

    def job_from_jenkins(self, jenkins_job: dict, project_root: str):
        """Create a Robota Job object from a Jenkins Job object."""
        self.name = jenkins_job["fullName"].replace(project_root + "/", "")
        self.short_name = jenkins_job["name"]
        self.url = jenkins_job["url"]
        if jenkins_job["lastBuild"]:
            self.last_build_number = jenkins_job["lastBuild"]["number"]
        if jenkins_job["lastCompletedBuild"]:
            self.last_completed_build_number = jenkins_job["lastCompletedBuild"]["number"]
        if jenkins_job["lastSuccessfulBuild"]:
            self.last_successful_build_number = jenkins_job["lastSuccessfulBuild"]["number"]

        for jenkins_build in jenkins_job["builds"]:
            self._builds.append(Build(jenkins_build))

    def get_builds(self) -> List[Build]:
        """Get all builds of a job."""
        return self._builds

    def get_build_by_number(self, number) -> Union[Build, None]:
        """Get build of this job by number, where 1 is the chronologically earliest build of a job.
        If build is not found, returns None."""
        for build in self._builds:
            if build.number == number:
                return build
        return None

    def get_last_completed_build(self) -> Union[Build, None]:
        """"Get the last completed build of a job."""
        try:
            return self.get_build_by_number(self.last_completed_build_number)
        except AttributeError:
            return None

    def get_last_build(self, start: datetime, end: datetime) -> Union[None, Build]:
        """Get most recent job build status between start and end.

        :param start: Build must occur after this time
        :param end: Build must occur before this time
        :return: Last build in time window, None if no job in time window.
        """
        if start is None or end is None:
            raise TypeError

        # Start with most recent build, looking for the last build before *end*
        builds = self.get_builds()
        for build in builds:
            if start < build.timestamp < end:
                return build
        return None

    def get_first_successful_build(self, start: datetime, end: datetime) -> Union[None, Build]:
        """Return the first (oldest) successful build in the time window."""
        for build in reversed(self.get_builds()):
            if start < build.timestamp < end and build.result == BuildResult.Success:
                return build
        return None

    def get_first_build(self, start: datetime, end: datetime) -> Union[None, Build]:
        """Return the first (oldest) build in the time window."""
        for build in reversed(self.get_builds()):
            if start < build.timestamp < end:
                return build
        return None

    def get_build_by_commit_id(self, commit_id) -> Union[Build, None]:
        """Get a job triggered by commit_id"""
        builds = self.get_builds()
        for build in builds:
            if commit_id == build.commit_id:
                return build
        return None


class CIServer(ABC):
    """A CIServer is a service from which test results are fetched. All of these are abstract
    methods implemented by subclasses.
    """
    def __init__(self):
        self._jobs: List[Job] = []
        self.tests: Dict[str, List[Test]] = {}

    @abstractmethod
    def get_jobs_by_folder(self, folder_name: str) -> List[Job]:
        """Get all jobs located in a particular folder."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_job_by_name(self, job_name: str) -> Union[Job, None]:
        """Get a job by its name. Return None if job not found."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_tests(self, job_path: str) -> Union[None, List[Test]]:
        """Get all Tests that were run for a job."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_package_coverage(self, job_path: str, package_name: str) -> Union[None, float]:
        """Get the percentage test coverage for a particular package."""
        raise NotImplementedError("Not implemented in base class.")


class JenkinsCIServer(CIServer):
    """With Jenkins it is possible to download all of the jobs from a whole project at once.
    This is much quicker than getting each job one by one as the API requests are slow. For this
    reason the JenkinsCIServer class downloads all jobs from a project and then helper methods
    get jobs from the local cache."""
    def __init__(self, ci_source: dict):
        """Connects to Jenkins and downloads all jobs. If the jobs are heavily nested in folders,
        it may be necessary to increase the depth parameter to iteratively fetch the lower level
        jobs.

        :param ci_source: A dictionary of config info for setting up the JenkinsCIServer.
        """
        super().__init__()
        self.url = ci_source["url"]
        token = ci_source["token"]
        username = ci_source["username"]

        self.project_name = ci_source["project_name"]
        self.folder_name = ci_source["folder_name"]
        self.base_request_string = f"{self.url}job/{self.project_name}/job/{self.folder_name}/"

        logger.info("Logging in to Jenkins to get CI information.")
        self.server = jenkins.Jenkins(self.url, username=username, password=token)

        # Populate the CIServer object with Jobs
        request_string = self._build_request_string(folder_depth=4)
        job_data = self.server.jenkins_open(requests.Request('GET', request_string))
        all_jobs = json.loads(job_data)
        self._populate_jobs(all_jobs)

    def _populate_jobs(self, nested_jobs):
        """Iteratively unfolds jobs from any containing folders, and stores all jobs as a
        flat list."""
        for child in nested_jobs["jobs"]:
            if child["_class"].endswith("Folder"):
                self._populate_jobs(child)
            else:
                self._add_job(child)

    def _add_job(self, jenkins_job: dict):
        """Adds a single job to the list of jobs in the CIJobServer instance."""
        self._jobs.append(Job(jenkins_job, f"{self.project_name}/{self.folder_name}"))

    def get_jobs_by_folder(self, folder_name: str) -> List[Job]:
        """Get all jobs that were located in a particular folder."""
        jobs = []
        for job in self._jobs:
            if job.name.startswith(folder_name):
                jobs.append(job)
        return jobs

    def get_job_by_name(self, job_name: str) -> Union[Job, None]:
        """Get a job by its name. Return None if job not found."""
        for job in self._jobs:
            if job.name == job_name:
                return job
        return None

    def _build_request_string(self, folder_depth=4) -> str:
        """Returns the request string for all of the Jenkins build results in a folder.
        The string is formed recursively since the jobs may be in nested folders.

        :param folder_depth: The number of folders deep to nest the xtree request.
        """
        jobs = "jobs[fullName,name,url,lastBuild[number],lastCompletedBuild[number]," \
               "lastSuccessfulBuild[number],BUILDS,JOBS]"
        builds = "builds[number,result,timestamp,url,actions" \
                 "[_class,lastBuiltRevision[SHA1,branch[*]],instructionCoverage[*]," \
                 "foundFailureCauses[*]]]"

        tree_string = jobs

        for i in range(folder_depth):
            tree_string = tree_string.replace("JOBS", jobs)
            if i == (folder_depth - 1):
                tree_string = tree_string.replace(",JOBS", "")
        tree_string = tree_string.replace("BUILDS", builds)
        return f"{self.base_request_string}/api/json?depth={folder_depth}&tree={tree_string}"

    def get_tests(self, job_path: str) -> Union[None, List[Test]]:
        """Get Tests for a job - this is a separate API request to the main job info."""
        if job_path in self.tests:
            return self.tests[job_path]

        job_name = job_path.replace('/', '/job/')
        job_name = f'job/{job_name}'
        request_string = f"{self.base_request_string}{job_name}/lastCompletedBuild/testReport/" \
                         f"api/json?tree=suites[cases[name,status],name,timestamp]"
        response = self._jenkins_get(request_string)
        if not response:
            return None
        data = json.loads(response)
        tests = self._process_test_names(data)
        self.tests[job_path] = tests

        return tests

    def get_package_coverage(self, job_path: str, package_name: str) -> Union[None, float]:
        """Get the percentage test coverage for a particular package.

        :param job_path: The tag or job name to query.
        :param package_name: The name of the package to get coverage for.
        """
        job_name = job_path.replace('/', '/job/')
        job_name = f'job/{job_name}'
        request_string = f"{self.base_request_string}{job_name}/lastCompletedBuild/jacoco/" \
                         f"{package_name}/api/json?tree=instructionCoverage[percentageFloat]"
        response = self._jenkins_get(request_string)
        if response is None:
            return None
        coverage = json.loads(response)
        return coverage["instructionCoverage"]["percentageFloat"]

    def _jenkins_get(self, request_string: str) -> Union[None, str]:
        """Send a direct API request to the open Jenkins server.

        :param request_string: The API request string to send.
        """
        request = requests.Request('GET', request_string)
        try:
            response = self.server.jenkins_open(request)
        except jenkins.NotFoundException:
            # If the job has not generated data corresponding to the request string
            # then the API request will fail.
            return None
        return response

    @staticmethod
    def _process_test_names(data: dict) -> List[Test]:
        """Get a list of test names from the nested JSON in the test report."""
        tests = []
        for suite in data["suites"]:
            for case in suite["cases"]:
                tests.append(Test(suite, case))
        return tests


# This type refers to any of the subclasses of CIServer - it is used for typing the return of the
# CIServer factory method.
CIType = TypeVar('CIType', bound=CIServer)


def new_ci_server(robota_config: dict) -> Union[None, CIType]:
    """Factory method for creating CIServers"""
    ci_server_source = config_readers.get_data_source_info(robota_config, 'ci')
    if not ci_server_source:
        return None
    ci_type = ci_server_source["type"]

    logger.debug(f"Initialising {ci_type} ci server.")

    if ci_server_source["type"] == 'jenkins':
        return JenkinsCIServer(ci_server_source)
    else:
        raise TypeError(f"Unknown CI server type {ci_server_source['type']}.")



import datetime
from typing import List, Union

from robota_core.commit import Commit


def is_date_before_other_dates(query_date: datetime.datetime, deadline: datetime.datetime,
                               key_date: datetime.datetime) -> bool:
    """Determine whether an action was before the deadline and another key date.

    :param query_date: The date of the action in question, e.g. when was an issue assigned, time
      estimate set, or due date set.
    :param deadline: The deadline of the action
    :param key_date: The query date should be before this significant date, as well as the
      deadline e.g. branch creation date
    :return: True if issue query_date was before deadline and key_date else False.
    """
    if query_date is not None and key_date is not None:
        assert key_date is not query_date

    if not query_date:
        return False

    if query_date < deadline:
        if not key_date:
            return True
        elif query_date < key_date:
            return True
        else:
            return False
    else:
        return False


def get_first_feature_commit(base_commits: List[Commit],
                             feature_commits: List[Commit]) -> Union[Commit, None]:
    """Get the first commit on a feature branch. Determine first commit by looking for
    branching points, described by the parent commits.
    All parameters are lists of commit IDs ordered from newest to oldest

    :param base_commits: commits from base branch (usually master)
    :param feature_commits: commits from feature branch
    :return first_feature_commit: commit ID of first commit on feature
    """

    # No feature branch commits in specified time range
    if not feature_commits:
        return None

    # To be certain that we have the first commit on feature,
    # there must be a commit common to both lists.
    if feature_commits[-1] != base_commits[-1]:
        raise AssertionError('The oldest commit in each list must be common to both branches.')

    # First check for unmerged feature branch
    # If last feature commit is not in base commits then the feature branch is unmerged
    if feature_commits[0] not in base_commits:
        # Now loop through commits and look at their parents.
        # The parent that is in base is the branching point
        for commit in feature_commits:
            if Commit({"id": commit.parent_ids[0]}, "dict") in base_commits:
                # If a parent is found in base then the commit is the first on the feature branch.
                return commit
        raise AssertionError("Feature branch not connected to master branch.")

    # Feature commit is in both feature and base, so feature parent must be
    # parent to a base commit too. Test for merged feature by looking for merge commit:
    for feature_commit in feature_commits:
        if find_feature_parent(feature_commit, base_commits):
            return feature_commit

    # All feature commits have been tested and neither an unmerged feature
    # branch, nor a merged feature with merge commit were found.
    # This is a FAST-FORWARD MERGE, so return the second commit on the feature branch
    # (the two lists of feature branch commits contains one extra,
    # earlier commit each so that we can always find a common commit).
    return feature_commits[-2]


def find_feature_parent(feature_commit: Commit, base_commits: List[Commit]) -> bool:
    """Determine whether the provided feature commit has a commit in the base branch with a
    common parent.

    :param feature_commit: The feature commit being checked.
    :param base_commits: A list of the base commits, most recent first.
    :returns: True if feature_commit has a common parent with a commit in the
      base branch else False.
    """
    base_commit = base_commits[0]
    while True:
        try:
            feature_parent = feature_commit.parent_ids[0]
        except IndexError:
            feature_parent = None

        if feature_parent in base_commit.parent_ids and base_commit != feature_commit:
            return True

        # The first parent is always the branch being merged into - the base branch
        try:
            base_commit = find_commit_in_list(base_commit.parent_ids[0], base_commits)
        except IndexError:
            base_commit = None

        if not base_commit:
            # If we have gone through all of the base commits and not found feature_parent then
            # this feature commit is not the first in the feature branch
            return False


def find_commit_in_list(commit_id: str, commits: List[Commit]) -> Union[None, Commit]:
    """Find a Commit in a list of Commits by its ID.

    :param commit_id: The id of the commit to find.
    :param commits: The list of Commits to search.
    :return: Commit if found, else None.
    """
    for commit in commits:
        if commit.id == commit_id:
            return commit
    return None


def fixup_first_feature_commit(feature_branch_commits: List[Commit],
                               initial_guess_of_first_commit: Commit, merge_commits: List[Commit]):
    """Fix-up function to look for merge commits on master branch before the tip of the feature
    branch. Any commits up to and including a merge commit in the history of a feature branch
    cannot be the first commit on the feature branch.

    :param feature_branch_commits:
    :return:
    """
    # next_commit means next chronologically, rather than next in the list of commits,
    # which is ordered newest to oldest.
    branch_tip = feature_branch_commits[0]
    next_commit = None
    for commit in feature_branch_commits:
        if commit in merge_commits and commit is not branch_tip:
            if next_commit:
                return next_commit
            else:
                return commit
        next_commit = commit
    else:
        return initial_guess_of_first_commit


def date_is_before(date1: datetime.datetime, date2: datetime.datetime) -> bool:
    """If date1 and date2 are provided and date1 is before date2 return True, else return False."""
    if date1 and date2 and date1 < date2:
        return True
    else:
        return False


def logical_and_lists(list1: List[bool], list2: List[bool]) -> List[bool]:
    """For two lists of booleans of length N, return a list of length N
    where output[i] is True if list1[i] is True and list2[i] is True,
    else output[i] is False.
    """
    assert len(list1) == len(list2)

    return [a and b for a, b in zip(list1, list2)]


def are_list_items_in_other_list(reference_list: List, query_list: List) -> List[bool]:
    """Check whether items in query_list exist in correct_list.

    :param reference_list: The reference list of items
    :param query_list: The items to check - does this list contain items in reference_list?
    :return items_present: Whether items in the correct list are in query_list (bool)

    >>> are_list_items_in_other_list([1, 2, 3], [3, 1, 1])
    [True, False, True]
    """
    items_present = [True if item in query_list else False for item in reference_list]

    return items_present


def are_lists_equal(list_1: list, list_2: list) -> List[bool]:
    """Elementwise comparison of lists. Return list of booleans, one for each element in the
    input lists, True if element N in list 1 is equal to element N in list 2, else False."""
    return [i == j for i, j in zip(list_1, list_2)]


def fraction_of_lists_equal(list_1: list, list_2: list) -> float:
    """Returns the fraction of list elements are equal when compared elementwise."""
    boolean_equals = are_lists_equal(list_1, list_2)
    return boolean_equals.count(True) / len(boolean_equals)


def get_value_from_list_of_dicts(list_of_dicts: List[dict], search_key: str, search_value: int,
                                 return_key: str):
    """Given a list of dictionaries, identify the required dictionary which contains the
    *search_key*: *search_value* pair. Return the value in that dictionary associated
    with *return_key*."""
    assert(search_key != return_key)
    for d in list_of_dicts:
        for k in d:
            if k == search_key and d[k] == search_value:
                return d[return_key]



"""Objects and for describing and processing Git commits."""

import datetime
import copy
from typing import List, Union, TYPE_CHECKING

import dateparser
import dateutil.parser
import gitlab.v4.objects
import github.Tag
import github.Commit
import github.CommitComment
import github.GithubObject
import git

from robota_core.string_processing import clean, get_link

if TYPE_CHECKING:
    from robota_core.repository import Event


class Commit:
    """An abstract object representing a git commit.

    :ivar created_at: (datetime) Commit creation time.
    :ivar id: Commit id.
    :ivar parent_ids: (List[string]) The ids of one or more commit parents.
    :ivar raw_message: (str) The original commit message
    :ivar message: (str) The commit message cleaned for HTML display.
    :ivar merge_commit: (bool) Whether this commit is a merge commit.
    """
    def __init__(self, commit, commit_source: str, project_url: str = None):
        self.created_at = None
        self.id = None
        self.author_name = None    # Used only for progress report
        self.short_id = None
        self.parent_ids = None
        self.raw_message = ""
        self.email = None
        self.comments: List["CommitComment"] = []
        self.url = None
        self.link = None
        self.network_url = None
        self.network_link = None

        if commit_source == "gitlab":
            self._commit_from_gitlab(commit, project_url)
        elif commit_source == "github":
            self._commit_from_github(commit)
        elif commit_source == "local":
            self.commit_from_local(commit)
        elif commit_source == "dict":
            self._commit_from_dict(commit)
        else:
            raise TypeError(f"Unknown commit type '{commit_source}'")

        self.message = clean(self.raw_message)
        self.merge_commit = self._is_merge_commit()

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return f"{self.short_id}"

    def get_comments(self) -> List[str]:
        """Return the text of all comments to this commit."""
        return [comment.text for comment in self.comments]

    def __eq__(self, other_commit: Union[None, "Commit"]):
        if other_commit is None:
            return False
        elif self.id == other_commit.id:
            return True
        else:
            return False

    def _commit_from_github(self, github_commit: github.Commit.Commit):
        commit = github_commit.commit

        self.created_at = dateutil.parser.parse(commit.last_modified)
        self.id = commit.sha
        self.author_name = commit.author.name
        self.short_id = self.id[:10]
        self.parent_ids = [parent.sha for parent in commit.parents]
        self.raw_message = commit.message
        self.email = commit.author.email
        comments = github_commit.get_comments()
        self.comments = [CommitComment(comment, "gitlab") for comment in comments]
        self.url = commit.html_url
        self.link = get_link(self.url, self.short_id)

    def _commit_from_gitlab(self, gitlab_commit: gitlab.v4.objects.ProjectCommit, project_url: str):
        """Convert a Gitlab commit to RoboTA Commit."""
        self.created_at = dateparser.parse(gitlab_commit.attributes["created_at"])
        self.id = gitlab_commit.attributes["id"]
        self.author_name = gitlab_commit.attributes["author_name"]
        self.short_id = gitlab_commit.attributes["short_id"]
        self.parent_ids = gitlab_commit.attributes["parent_ids"]
        self.raw_message = gitlab_commit.attributes["message"]
        self.email = gitlab_commit.attributes['author_email'].lower()
        comments = gitlab_commit.comments.list(all=True)
        self.comments = [CommitComment(comment, "gitlab") for comment in comments]
        self.url = f'{project_url}/commit/{self.id}'
        self.link = get_link(self.url, self.short_id)
        self.network_url = f'{project_url}/network/master?utf8=✓&extended_sha1={self.id}'
        self.network_link = get_link(self.network_url, self.short_id)

    def commit_from_local(self, commit: git.Commit):
        self.created_at = commit.authored_datetime
        self.id = commit.hexsha
        self.author_name = commit.author.name
        self.short_id = self.id[:10]
        self.parent_ids = [parent.hexsha for parent in commit.parents]
        self.raw_message = commit.message
        self.email = commit.author.email

    def _commit_from_dict(self, commit: dict):
        """Used for testing, create a commit with just the ID and ID of parents."""
        self.id = commit["id"]
        if "parents" in commit:
            self.parent_ids = commit["parents"]
        else:
            self.parent_ids = []

    def _is_merge_commit(self) -> bool:
        """Is this a merge commit? Only a merge commit can have more than one parent.

        :return: Whether this is a merge commit
        """
        if len(self.parent_ids) > 1:
            return True
        else:
            return False


class CommitComment:
    """A comment made on a commit."""
    def __init__(self, comment_data, source: str):
        self.text = None
        self.author = None

        if source == "gitlab":
            self._commit_comment_from_gitlab(comment_data)
        elif source == "github":
            self._commit_comment_from_github(comment_data)
        else:
            raise TypeError("Unknown commit comment data type.")

    def _commit_comment_from_gitlab(self, comment_data: gitlab.v4.objects.ProjectCommitComment):
        self.text = clean(comment_data.attributes["note"])
        self.author = comment_data.attributes['author']

    def _commit_comment_from_github(self, comment_data: github.CommitComment.CommitComment):
        self.text = comment_data.body
        self.author = comment_data.user.name


class Tag:
    """A tag is a named pointer to a git commit.

    :ivar name: The name of the tag.
    :ivar commit_id: The id of the commit that the tag points to.
    """
    def __init__(self, tag_data, source: str):
        self.name = ""
        self.commit_id = ""

        if source == "gitlab":
            self.tag_from_gitlab(tag_data)
        elif source == "github":
            self.tag_from_github(tag_data)
        elif source == "local":
            self.tag_from_local(tag_data)
        elif source == "dict":
            self.tag_from_dict(tag_data)
        else:
            raise TypeError(f"Cannot create tag, unknown tag data type: {source}")

    def tag_from_gitlab(self, tag_data: gitlab.v4.objects.ProjectTag):
        self.name = tag_data.attributes["name"]
        self.commit_id = tag_data.attributes["commit"]["id"]

    def tag_from_dict(self, tag_data: dict):
        self.name = tag_data["name"]
        self.commit_id = tag_data["commit_id"]

    def tag_from_local(self, tag_data: git.Tag):
        self.name = tag_data.name
        self.commit_id = tag_data.commit.hexsha

    def tag_from_github(self, tag_data: github.Tag.Tag):
        self.name = tag_data.name
        self.commit_id = tag_data.commit.sha


class CommitCache:
    """A cache of Commit objects from a specific date range and branch."""
    def __init__(self, start: datetime.datetime, end: datetime.datetime, branch: str,
                 commits: List[Commit]):
        self.start = start
        self.end = end
        self.branch = branch
        self.commits = tuple(commits)

    def __iter__(self):
        yield from self.commits


def get_merge_commit(feature_tip: Commit, master_commits: List[Commit]) -> Union[Commit, None]:
    """Get merge commit ID for the branch "branch_title".
    Given the id of the commit at the tip of a feature branch, find where it merges into the
    master branch by going through the commits ids on the master branch and looking at their
    parents.

    :param feature_tip: The Commit at the tip of the feature branch.
    :param master_commits: Commits of master branch, ordered by date, most recent first.
    :return merge_commit: The id of the merge commit if branch was merged else returns None.
    """
    if not master_commits:
        return None

    assert isinstance(feature_tip, Commit)

    # Find where the branch tip is in the list of master Commits
    master_commit_index = None
    for index, commit in enumerate(master_commits):
        if commit == feature_tip:
            master_commit_index = index
            break
    if master_commit_index is None:
        # Branch was not merged
        return None

    # Search for merge commits in the more recent commits.
    for commit in reversed(master_commits[:master_commit_index]):
        if len(commit.parent_ids) > 1 and feature_tip.id == commit.parent_ids[1]:
            # Non-FF merge
            return commit
    # FF merge
    return feature_tip


def get_tags_at_date(date: datetime.datetime, tags: List[Tag],
                     events: List["Event"]) -> List[Tag]:
    tags = copy.deepcopy(tags)

    for event in events:
        if event.date > date:
            # Add tags that have been deleted since date.
            if event.type == "deleted":
                if event.ref_type == "tag":
                    tag_data = {"name": event.ref_name,
                                "commit_id": event.commit_id}
                    tags.append(Tag(tag_data, "dict"))

            # Remove tags that have been added since date.
            elif event.type == "pushed to" or event.type == "pushed new":
                if event.ref_type == "tag":
                    tag_name = event.ref_name
                    tag_commit = event.commit_id
                    pointer = 0
                    while pointer < len(tags):
                        if tags[pointer].commit_id == tag_commit and tags[pointer].name == tag_name:
                            del (tags[pointer])
                        else:
                            pointer += 1
    return tags



import pathlib

import sys

from loguru import logger


class RemoteProviderError(Exception):
    """The error raised when there is a problem with data from a remote provider."""


def set_up_logger():
    log_path = pathlib.Path("robota.log")
    try:
        log_path.unlink()
    except FileNotFoundError:
        pass

    # remove the default sink before adding new ones
    logger.remove()
    logger.add(sys.stderr, format="{level} - {message}", level="SUCCESS")
    logger.add("robota.log", format="{time:YYYY-MM-DD HH:mm:ss}: {level} - {message}",
               level="DEBUG", delay=True)


set_up_logger()



import datetime
import io
import sys
from abc import abstractmethod
from typing import List, Union

import github
import github.Branch
from github.File import File
import gitlab
import gitlab.v4.objects
import git
from loguru import logger

from robota_core import gitlab_tools, config_readers
from robota_core.commit import CommitCache, Tag, Commit, get_tags_at_date
from robota_core.github_tools import GithubServer
from robota_core.string_processing import string_to_datetime


class Branch:
    """An abstract object representing a git branch.

    :ivar id: Name of branch.
    :ivar id: Commit id that branch points to.
    """
    def __init__(self, branch, source: str):
        self.name = None
        self.id = None

        if source == "gitlab":
            self._branch_from_gitlab(branch)
        elif source == "github":
            self._branch_from_github(branch)
        elif source == "dict":
            self._branch_from_dict(branch)
        elif source == "local":
            self._branch_from_local(branch)
        else:
            TypeError("Unknown branch type.")

    def _branch_from_gitlab(self, branch: gitlab.v4.objects.ProjectBranch):
        self.name = branch.attributes["name"]
        self.id = branch.attributes["commit"]["id"]

    def _branch_from_dict(self, branch: dict):
        self.name = branch["name"]
        self.id = branch["commit_id"]

    def _branch_from_github(self, branch: github.Branch.Branch):
        self.name = branch.name
        self.id = branch.commit.sha

    def _branch_from_local(self, branch: git.Head):
        self.name = branch.name
        self.id = branch.commit.hexsha


class Event:
    """A repository event.

    :ivar date: The date and time of the event.
    :ivar type: 'deleted', 'pushed to' or 'pushed new'
    :ivar ref_type: The thing the event concerns, 'tag', 'branch', 'commit' etc.
    :ivar ref_name: The name of the thing the event concerns (branch name or tag name)
    :ivar commit_id: A commit id associated with ref
    """
    def __init__(self, event_data):
        self.date = None
        self.type = None
        self.ref_type = None
        self.ref_name = None
        self.commit_id = None
        self.commit_count = None

        if isinstance(event_data, gitlab.v4.objects.ProjectEvent):
            self._event_from_gitlab(event_data)
        elif isinstance(event_data, dict):
            self._event_from_dict(event_data)

    def _event_from_gitlab(self, event_data: gitlab.v4.objects.ProjectEvent):
        self.date = string_to_datetime(event_data.attributes['created_at'])
        self.type = event_data.attributes["action_name"]

        if "push_data" in event_data.attributes:
            push_data = event_data.attributes['push_data']
            self.ref_type = push_data['ref_type']
            self.ref_name = push_data['ref']
            self.commit_count = push_data['commit_count']
            if self.type == "deleted":
                self.commit_id = push_data['commit_from']
            else:
                self.commit_id = push_data['commit_to']

    def _event_from_dict(self, event_data: dict):
        self.date = string_to_datetime(event_data['date'])
        self.type = event_data["type"]

        if "push_data" in event_data:
            push_data = event_data['push_data']
            self.ref_type = push_data['ref_type']
            self.ref_name = push_data['ref_name']
            self.commit_id = push_data['commit_id']
            self.commit_count = push_data['commit_count']


class Diff:
    """A representation of a git diff between two points in time for a single
    file in a git repository."""
    def __init__(self, diff_info, diff_source: str):
        self.old_path: str
        self.new_path: str
        self.new_file: bool
        self.diff: str

        if diff_source == "gitlab":
            self._diff_from_gitlab(diff_info)
        elif diff_source == "github":
            self._diff_from_github(diff_info)
        elif diff_source == "local_repo":
            self._diff_from_local(diff_info)
        else:
            raise TypeError(f"Unknown diff source: '{diff_source}'")

    def _diff_from_gitlab(self, diff_info: dict):
        """Populate a diff using the dictionary of diff information returned by gitlab."""
        self.old_path = diff_info["old_path"]
        self.new_path = diff_info["new_path"]
        self.new_file = diff_info["new_file"]
        self.diff = diff_info["diff"]

    def _diff_from_local(self, diff_info: git.Diff):
        self.old_path = diff_info.a_path
        self.new_path = diff_info.b_path
        self.new_file = diff_info.new_file
        self.diff = diff_info.diff

    def _diff_from_github(self, diff_info: github.File.File):
        self.new_path = diff_info.filename
        if diff_info.previous_filename is None:
            self.old_path = self.new_path
        else:
            self.old_path = diff_info.previous_filename
        if diff_info.status == "added":
            self.new_file = True
        else:
            self.new_file = False
        self.diff = diff_info.patch


class Repository:
    """A place where commits, tags, events, branches and files come from.

    :ivar _branches: A list of Branches associated with this repository.
    :ivar _events: A list of Events associated with this repository.
    :ivar _diffs: A dictionary of cached diffs associated with this repository. They are labelled
      in the form key = point_1 + point_2 where point_1 and point_2 are the commit SHAs or branch
      names that the diff describes.
    """
    def __init__(self, project_url: str):
        self._branches: Union[None, List[Branch]] = None
        self._events: List[Event] = []
        self._diffs = {}
        self._stored_commits: List[CommitCache] = []
        self._tags: List[Tag] = []
        self.project_url = project_url

    @abstractmethod
    def list_files(self, identifier: str) -> List[str]:
        """Returns a list of file paths with file names in a repository. Identifier can be a
        commit or branch name. File paths are relative to the repository root."""
        raise NotImplementedError("Not implemented in base class.")

    def get_branches(self) -> List[Branch]:
        """Get all of the Branches in the repository."""
        if not self._branches:
            self._branches = self._fetch_branches()
        return self._branches

    def get_branch(self, name: str) -> Union[Branch, None]:
        """Get a Branch from the repository by name. If Branch does not exist, return None."""
        if not self._branches:
            self._branches = self._fetch_branches()
        for branch in self._branches:
            if branch.name == name:
                return branch
        return None

    @abstractmethod
    def _fetch_branches(self) -> List[Branch]:
        """Retrieve all branches from the server."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_events(self) -> List[Event]:
        """Return a list of Events associated with this repository."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def get_file_contents(self, file_path: str, branch: str = "master") -> Union[bytes, None]:
        """Get the decoded contents of a file from the repository. Works well for text files. Might
        explode for other file types."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def compare(self, point_1: str, point_2: str) -> List[Diff]:
        """Compare the state of the repository at two points in time.
        The points may be branch names, tags or commit ids.
        """
        raise NotImplementedError("Not implemented in base class.")

    def get_commits(self, start: datetime.datetime = None, end: datetime.datetime = None,
                    branch: str = None) -> List[Commit]:
        """Get issues from the issue provider between the start date and end date."""
        cached_commits = self._get_cached_commits(start, end, branch)
        if cached_commits:
            return list(cached_commits.commits)

        new_commits = self._fetch_commits(start, end, branch)
        self._stored_commits.append(CommitCache(start, end, branch, new_commits))
        return new_commits

    def get_commit_by_id(self, commit_id: str) -> Union[Commit, None]:
        """Get a Commit by its unique ID number"""
        if commit_id is None:
            return None

        for cache in self._stored_commits:
            for commit in cache:
                if commit.id.startswith(commit_id):
                    return commit

        new_commit = self._fetch_commit_by_id(commit_id)

        # Add the new commit to a cache of its own.
        fake_date = datetime.datetime.fromtimestamp(1, datetime.timezone.utc)
        new_cache = CommitCache(fake_date, fake_date, "", [new_commit])
        self._stored_commits.append(new_cache)

        return new_commit

    def get_tags(self):
        """Get all tags from the server."""
        if not self._tags:
            self._tags = self._fetch_tags()
        return self._tags

    def get_tag(self, name: str, deadline: datetime.datetime = None,
                events: List["Event"] = None) -> Union[Tag, None]:
        """Get a git Tag by name.

        :param name: The name of the tag to get.
        :param deadline: If provided, filters tags such that tags are only returned if they
          existed at deadline.
        :param events: Events corresponding to the repository, required if deadline is specified.
        :returns: The Tag if found else returns None.
        """
        if not self._tags:
            self._tags = self._fetch_tags()
        tags_to_search = self._tags

        if deadline:
            if not events:
                raise SyntaxError("Must provide list of events if deadline is specified.")
            tags_to_search = get_tags_at_date(deadline, tags_to_search, events)

        for tag in tags_to_search:
            if tag.name == name:
                return tag
        return None

    def _get_cached_commits(self, start: datetime.datetime,
                            end: datetime.datetime, branch: str) -> Union[CommitCache, None]:
        """Check whether commits with the specified start, end and branch are already stored."""
        for cache in self._stored_commits:
            if cache.start == start and cache.end == end and cache.branch == branch:
                return cache
        return None

    @abstractmethod
    def _fetch_tags(self) -> List[Tag]:
        """Fetch tags from a the server."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def _fetch_commit_by_id(self, commit_id: str) -> Union[Commit, None]:
        """Fetch a single commit from the server."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def _fetch_commits(self, start: Union[datetime.datetime, None],
                       end: Union[datetime.datetime, None],
                       branch: Union[str, None]) -> List[Commit]:
        """Fetch a list of commits from the server."""
        raise NotImplementedError("Not implemented in base class.")


class LocalRepository(Repository):

    def __init__(self, commit_source: dict):
        super().__init__(commit_source["path"])
        self.repo = git.Repo(commit_source["path"])

    def list_files(self, identifier: str) -> List[str]:
        files = self.repo.tree(identifier).traverse()
        file_paths = [file.path for file in files if file.type == "blob"]
        return file_paths

    def get_file_contents(self, file_path: str, branch: str = "master") -> Union[bytes, None]:
        file = self.repo.heads[branch].commit.tree / file_path

        with io.BytesIO(file.data_stream.read()) as f:
            return f.read()

    def compare(self, point_1: str, point_2: str) -> List[Diff]:
        commit_1 = self._get_commit(point_1)
        commit_2 = self._get_commit(point_2)
        diffs = commit_1.diff(commit_2, create_patch=True)
        robota_diffs = [Diff(diff, "local_repo") for diff in diffs]
        return robota_diffs

    def _get_commit(self, ref: str) -> git.Commit:
        """Get a commit object from a ref which is a branch, tag or commit SHA."""
        try:
            commit = self.repo.heads[ref].commit
            return commit
        except IndexError:
            pass
        try:
            commit = self.repo.tags[ref].commit
            return commit
        except IndexError:
            pass
        try:
            commit = self.repo.commit(ref)
            return commit
        except IndexError:
            logger.error(f"Can't find object {ref} in Local repository. Object must be branch name,"
                         f"tag or commit SHA.")
            sys.exit(1)

    def get_events(self) -> List[Event]:
        # TODO: Can events be mined from the reflog?
        raise NotImplementedError("Get events not implemented for LocalRepository")

    def _fetch_branches(self) -> List[Branch]:
        return [Branch(branch, "local") for branch in self.repo.branches]

    def _fetch_commits(self, start: Union[datetime.datetime, None],
                       end: Union[datetime.datetime, None],
                       branch: Union[str, None]) -> List[Commit]:
        rev_list_args = {}
        if start:
            rev_list_args["since"] = start.isoformat()
        if end:
            rev_list_args["until"] = end.isoformat()
        rev = None
        if branch:
            rev = branch
        commits = self.repo.iter_commits(rev, "", **rev_list_args)
        return [Commit(commit, "local") for commit in commits]

    def _fetch_commit_by_id(self, commit_id: str) -> Union[Commit, None]:
        return Commit(self.repo.commit(commit_id), "local")

    def _fetch_tags(self) -> List[Tag]:
        return [Tag(tag, "local") for tag in self.repo.tags]


class GithubRepository(Repository):
    def __init__(self, repository_source: dict):
        super().__init__(repository_source['url'])
        server = GithubServer(repository_source)
        self.repo = server.open_github_repo(repository_source["project"])

    def list_files(self, identifier: str) -> List[str]:
        files = self.repo.get_git_tree(identifier, recursive=True)
        file_paths = [file.path for file in files.tree if file.type == "blob"]
        return file_paths

    def _fetch_branches(self) -> List[Branch]:
        return [Branch(branch, "github") for branch in self.repo.get_branches()]

    def get_events(self) -> List[Event]:
        raise NotImplementedError("Method not implemented for Github Repository")

    def get_file_contents(self, file_path: str, branch: str = "master") -> Union[bytes, None]:
        try:
            file = self.repo.get_contents(file_path, branch)
        except github.UnknownObjectException:
            return None
        return file.decoded_content

    def compare(self, point_1: str, point_2: str) -> List[Diff]:
        comparison = self.repo.compare(point_1, point_2)
        return [Diff(diff, "github") for diff in comparison.files]

    def _fetch_commit_by_id(self, commit_id: str) -> Union[Commit, None]:
        try:
            commit_data = self.repo.get_commit(commit_id)
        except github.GithubException as e:
            if e.status == 422:
                return None
            else:
                raise e

        return Commit(commit_data, "github")

    def _fetch_commits(self, start: Union[datetime.datetime, None],
                       end: Union[datetime.datetime, None],
                       branch: Union[str, None]) -> List[Commit]:
        if not start:
            start = github.GithubObject.NotSet
        if not end:
            end = github.GithubObject.NotSet
        if not branch:
            branch = github.GithubObject.NotSet

        github_commits = self.repo.get_commits(sha=branch, since=start, until=end)
        return [Commit(github_commit, "github") for github_commit in github_commits]

    def _fetch_tags(self) -> List[Tag]:
        github_tags = self.repo.get_tags()
        return [Tag(github_tag, "github") for github_tag in github_tags]


class GitlabRepository(Repository):
    """A Gitlab flavour of a repository.

    :ivar project: A connection to the gitlab repository
    """

    def __init__(self, data_source: dict):
        if "token" in data_source:
            token = data_source["token"]
        else:
            token = None
        server = gitlab_tools.GitlabServer(data_source["url"], token)
        self.project = server.open_gitlab_project(data_source["project"])

        super().__init__(self.project.attributes["web_url"])

    def list_files(self, identifier: str) -> List[str]:
        files = []
        page_num = 1
        while True:
            file_page = self.project.repository_tree(ref=identifier, per_page=100, page=page_num,
                                                     recursive=True)
            if file_page:
                files.extend(file_page)
                page_num += 1
            else:
                break
        file_paths = [file["path"] for file in files if file["type"] == "blob"]
        return file_paths

    def _fetch_branches(self) -> List[Branch]:
        return [Branch(branch, "gitlab") for branch in self.project.branches.list(all=True)]

    def get_events(self) -> List[Event]:
        """Return a list of Events associated with this repository."""
        if not self._events:
            # API requires date in ISO 8601 format
            gitlab_events = self.project.events.list(all=True, action="pushed")

            for gitlab_event in gitlab_events:
                self._events.append(Event(gitlab_event))
        return self._events

    def get_file_contents(self, file_path: str, branch: str = "master") -> Union[bytes, None]:
        """Get a file directly from the repository."""
        try:
            file = self.project.files.get(file_path, branch)
        except gitlab.GitlabGetError:
            return None
        else:
            return file.decode()

    def compare(self, point_1: str, point_2: str) -> List[Diff]:
        """Compare the state of the repository at two points in time.
        The points may be branch names, tags or commit ids.
        Point 1 must be chronologically before point 2.
        """
        if not point_1 + point_2 in self._diffs:
            gitlab_diffs = self.project.repository_compare(point_1, point_2)
            robota_diffs = [Diff(diff, "gitlab") for diff in gitlab_diffs["diffs"]]
            self._diffs[point_1 + point_2] = robota_diffs

        return self._diffs[point_1 + point_2]

    def _fetch_commits(self, start: Union[datetime.datetime, None],
                       end: Union[datetime.datetime, None],
                       branch: Union[str, None]) -> List[Commit]:
        """ Function to return commits falling withing a certain time window.

        :param start: The start of the time window for included commits.
        :param end: The end of the time window for included commits.
        :param branch: Filters commits by branch name. If None, get commits from all branches.
        :return: A list of Commit object.
        """
        request_parameters = {}

        if start is not None:
            request_parameters['since'] = start.isoformat()

        if end is not None:
            request_parameters['until'] = end.isoformat()

        if branch is None:
            request_parameters['all'] = True
        else:
            request_parameters['ref_name'] = branch

        gitlab_commits = self.project.commits.list(all=True,
                                                   query_parameters=request_parameters)

        return [Commit(commit,  "gitlab", self.project_url) for commit in gitlab_commits]

    def _fetch_commit_by_id(self, commit_id: str) -> Union[Commit, None]:
        try:
            gitlab_commit = self.project.commits.get(commit_id)
        except gitlab.exceptions.GitlabGetError:
            return None

        return Commit(gitlab_commit, "gitlab", self.project_url)

    def _fetch_tags(self) -> List[Tag]:
        """Method for getting tags from the gitlab server."""
        gitlab_tags = self.project.tags.list(all=True)
        return [Tag(gitlab_tag, "gitlab") for gitlab_tag in gitlab_tags]


def new_repository(robota_config: dict) -> Union[Repository, None]:
    """Factory method for Repositories."""
    repo_config = config_readers.get_data_source_info(robota_config, "repository")
    if not repo_config:
        return None
    repo_type = repo_config["type"]

    logger.debug(f"Initialising {repo_type} repository.")

    if repo_type == "gitlab":
        return GitlabRepository(repo_config)
    elif repo_type == "github":
        return GithubRepository(repo_config)
    elif repo_type == 'local_repository':
        return LocalRepository(repo_config)
    else:
        raise TypeError(f"Unknown repository type {repo_config['type']}.")



from loguru import logger
import os
import pathlib
import re
import shutil
import stat
import tempfile
import tarfile
import csv
from typing import List, Tuple, Union

import gitlab
import yaml

from robota_core import gitlab_tools as gitlab_tools


class RobotaConfigLoadError(Exception):
    """The error raised when there is a problem loading the configuration"""


class RobotaConfigParseError(Exception):
    """The error raised when there is a problem parsing the configuration"""


def get_config(file_names: Union[str, List[str]], data_source: dict) -> list:
    """The base method of the class. Calls different methods to get the config
    depending on the config type.

    :param file_names: The names of one or more config files to open.
    :param data_source: Information about the source of the config data. The 'type' key specifies
      the source of the data and other keys are extra information about the source like url
      or API token.
    :return: a list of parsed file contents, one list element for each file specified in
      `file_names`. If a file is not found, the corresponding list element is set to None.
    """
    if isinstance(file_names, str):
        file_names = [file_names]

    if not isinstance(data_source, dict):
        raise TypeError("Config variables must be a dictionary of variables.")
    source_type = data_source["type"]

    if source_type == "local_path":
        parsed_variables = _config_from_local_path(data_source, file_names)
    elif source_type == "gitlab":
        parsed_variables = _config_from_gitlab(data_source, file_names)
    else:
        raise RobotaConfigLoadError(f"Source type: {source_type} is not a valid data source for"
                                    f" the 'get_config' method.")

    return parsed_variables


def _config_from_gitlab(data_source, file_names) -> List[dict]:
    parsed_variables = []

    config_file_directory, commit_id = get_gitlab_config(data_source)
    for name in file_names:
        config_path = config_file_directory / pathlib.Path(name)
        if config_path.is_file():
            file_contents = parse_config(config_path)
            if isinstance(file_contents, dict):
                file_contents["config_commit_id"] = commit_id
            parsed_variables.append(file_contents)
        else:
            parsed_variables.append(None)
    if config_file_directory:
        shutil.rmtree(config_file_directory, onerror=rmtree_error)
    return parsed_variables


def _config_from_local_path(data_source, file_names) -> List[dict]:
    parsed_variables = []

    for name in file_names:
        config_path = pathlib.Path(data_source["path"]) / pathlib.Path(name)
        if config_path.exists():
            config = parse_config(config_path)
            parsed_variables.append(config)
        else:
            logger.warning(f"Attempted to load config from path: '{config_path}', "
                           f"but path does not exist.")
            parsed_variables.append(None)
    return parsed_variables


def get_gitlab_config(config_variables: dict) -> Tuple[pathlib.Path, str]:
    """Get config from a Gitlab repository by logging in using an access token and downloading
    the files from the repository.

    :param config_variables: required keys/value pairs are:
      gitlab_url: The URL  of the gitlab repository
      gitlab_project: The full project name of the project containing the config files.
      gitlab_token: The gitlab access token.
    :return: the temporary directory containing the config files.
    """
    # Read GitLab authentication token stored as environment variable
    if "token" in config_variables:
        token = config_variables["token"]
    else:
        token = None

    gitlab_server = gitlab_tools.GitlabServer(config_variables["url"], token)
    if "branch" in config_variables:
        branch_name = config_variables["branch"]
    else:
        branch_name = "master"

    project = gitlab_server.open_gitlab_project(config_variables["project"])
    try:
        commit_id = project.commits.get(branch_name).attributes["short_id"]
    except gitlab.exceptions.GitlabGetError:
        raise gitlab.exceptions.GitlabGetError(f"Error fetching config. "
                                               f"Branch {branch_name} not found in "
                                               f"repository: {config_variables['project']}.")
    temp_path = pathlib.Path(tempfile.mkdtemp())
    tar_path = temp_path / pathlib.Path("tar")
    with open(tar_path, 'wb') as output_dir:
        output_dir.write(project.repository_archive(branch_name))
    with tarfile.TarFile.open(tar_path, mode='r') as input_tar:
        input_tar.extractall(temp_path)
    for file in temp_path.iterdir():
        if file.name != "tar":
            return file, commit_id


def read_csv_file(csv_path: Union[str, pathlib.Path]) -> dict:
    """Parse a two column csv file. Return dict with first column as keys and second column
    as values.
    """
    data = {}
    with open(csv_path, newline='') as f:
        reader = csv.reader(f, skipinitialspace=True)
        for row in reader:
            if row:
                data[row[0]] = row[1]

    return data


def process_yaml(yaml_content: dict) -> dict:
    """Do custom string substitution to the dictionary produced from reading a YAML file.
    This is not part of the core YAML spec.
    This function replaces instances of ${key_name} in strings nested as values in dicts or lists
    with the value of the key "key_name" if "key_name" occurs in the root of the dictionary.
    """

    if isinstance(yaml_content, dict):
        # Collect all of the highest level values in the dict - these can be used for substitution
        # elsewhere
        root_keys = {}
        for key, value in yaml_content.items():
            if not isinstance(value, list) and not isinstance(value, dict):
                root_keys.update({key: value})

        for key, value in yaml_content.items():
            yaml_content[key] = substitute_dict(value, root_keys)
    return yaml_content


def substitute_dict(input_value: object, root_keys: dict) -> object:
    """If `input_value` is a list or dict, recurse into it trying to find strings.
    If `input_value` is a string then substitute any variables that occur as keys in `root_keys`
    with the values in `root_keys`.
    Variables to be substituted are indicated by a bash like syntax, e.g. ${variable_name}."""

    if isinstance(input_value, list):
        for item in input_value:
            input_value[input_value.index(item)] = substitute_dict(item, root_keys)
    if isinstance(input_value, dict):
        for key, value in input_value.items():
            input_value[key] = substitute_dict(value, root_keys)
    if isinstance(input_value, str):
        sub_keys = re.findall(r"(?<=\${)([^}]*)(?=})", input_value)
        for key in sub_keys:
            if key in root_keys:
                input_value = input_value.replace(f"${{{key}}}", str(root_keys[key]))
    return input_value


def parse_config(config_path: pathlib.Path) -> dict:
    """Parses a config file to extract the configuration variables from it.

    :param config_path: the full file path to the config file.
    :return: the config variables read from the file. Return type depends on the file type.
    """
    config_file_type = config_path.suffix

    if config_file_type in [".yaml", ".yml"]:
        config = read_yaml_file(config_path)
        config = process_yaml(config)
    elif config_file_type == ".csv":
        config = read_csv_file(config_path)
    else:
        raise RobotaConfigParseError(f"Cannot parse file of type: {type}.")

    return config


def read_yaml_file(config_location: pathlib.Path) -> dict:
    """ Read a YAML file into a dictionary

    :param config_location: the path of the config file
    :return: Key-value pairs from the config file
    """
    # noinspection PyTypeChecker
    with open(config_location, encoding='utf8') as yaml_file:
        try:
            config = yaml.load(yaml_file.read(), Loader=yaml.FullLoader)
        except yaml.YAMLError as e:
            logger.error(f"YAML Parsing of file {config_location.absolute()} failed.")
            raise e
    return config


def get_robota_config(config_path: str, substitution_vars: dict) -> dict:
    """The robota config specifies the source for each data type used by RoboTA. The RoboTA
    config is always stored locally since it contains API tokens.
    :param config_path: The path of the robota config file to read.
    :param substitution_vars: An optional dictionary of values to substitute into the config file.
    """
    config_path = pathlib.Path(config_path)

    # Load RoboTA config from file.
    robota_config = get_config([config_path.name], {"type": "local_path",
                                                    "path": config_path.parent})[0]
    if robota_config is None:
        raise RobotaConfigLoadError(f"Unable to load robota config from {config_path.absolute()}")
    logger.debug(f"robota-config loaded from {config_path.absolute()}")

    return substitute_keys(robota_config, substitution_vars)


def substitute_keys(robota_config: dict, command_line_args: dict) -> dict:
    """Go through all of the data sources replacing any curly bracketed strings by named
    variables provided to roboTA as command line arguments. This allows substitution of things
    like a team name or exercise number into the robota config.

    :param robota_config: The dictionary of data sources loaded from robota-config.yaml.
    :param command_line_args: Command line arguments given to RoboTA.
    """
    for top_key in ["data_types", "data_sources"]:
        for source_name, data_source in robota_config[top_key].items():
            for key, value in data_source.items():
                if not value:
                    raise KeyError(f"Key '{key}' in robota config has no value.")
                for name, arg in command_line_args.items():
                    if f"{{{name}}}" in value:
                        robota_config[top_key][source_name][key] = robota_config[top_key][source_name][key].replace(f"{{{name}}}", arg)

    return robota_config


def get_data_source_info(robota_config: dict, key: str) -> Union[dict, None]:
    """Get the information about the data source specified by 'key' from the robota_config
    dictionary."""
    config_error = "Error in RoboTA config file."
    if "data_types" not in robota_config:
        raise RobotaConfigParseError(f"{config_error} 'data_types' section not found in "
                                     f"robota-config.")
    if key not in robota_config["data_types"]:
        logger.debug(f"'{key}' not found in 'data_types' config section. Not initialising "
                     f"this data source.")
        return None
    data_type_info = robota_config["data_types"][key]
    if "data_source" not in data_type_info:
        raise RobotaConfigParseError(f"{config_error} 'data_source' key not found in details "
                                     f"of '{key}' data type in robota_config.")
    data_source = data_type_info["data_source"]
    if "data_sources" not in robota_config:
        raise RobotaConfigParseError(f"{config_error} 'data_sources section not found.")
    if data_source not in robota_config["data_sources"]:
        raise RobotaConfigParseError(f"{config_error} Data source '{data_source}' specified in "
                                     f"'data_types' section, but no details provided in "
                                     f"'data_sources' section.")
    data_source_info = robota_config["data_sources"][data_source]
    if "type" not in data_source_info:
        raise RobotaConfigParseError(f"Error in RoboTA config file. 'type' not specified in "
                                     f"data source: '{data_source}'.")
    return {**data_source_info, **data_type_info}


def rmtree_error(func, path, _):
    """Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file) it attempts to add write
    permission and then retries.
    """
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)



import datetime
import re
from typing import Union, Dict

import bleach
import markdown


def string_to_datetime(date: Union[str, None],
                       datetime_format: str = '%Y-%m-%dT%H:%M:%S.%f%z') -> Union[datetime.datetime,
                                                                                None]:
    """Convert time string (output from GitLab project attributes) to datetime.

    :param date: A string representing the datetime.
    :param datetime_format: The format of 'date'.
    :return date: The converted datetime.

    >>> string_to_datetime('2017-12-06T08:28:32.000Z', "%Y-%m-%dT%H:%M:%S.%fZ")
    datetime.datetime(2017, 12, 6, 8, 28, 32)
    """
    if date is None:
        return None
    if isinstance(date, str):
        gmt_date = date.replace('Z', '+00:00')
        return datetime.datetime.strptime(gmt_date, datetime_format)
    else:
        raise TypeError("Unknown date type. Cannot convert.")


def markdownify(text: str) -> str:
    """Take text in markdown format and output the formatted text with HTML markup."""
    return markdown.markdown(text, extensions=['attr_list'])


def clean(text: str) -> str:
    """Convert any HTML tags to a string representation so HTML cannot be executed."""
    return bleach.clean(text)


def html_newlines(text: str) -> str:
    """Replace any newline characters in a string with html newline characters."""
    html = re.sub('(\n)+', '<br>', text)
    return html


def list_to_html_rows(list_of_strings: list) -> str:
    """Join list items with html new lines."""
    return '<br>'.join(list_of_strings)


def sublist_to_html_rows(list_of_lists: list, empty='-') -> list:
    """Separate items in a sub-list by html new lines instead of commas."""
    for item in list_of_lists:
        assert item is None or isinstance(item, list)

    return [list_to_html_rows(list_items) if list_items else empty
            for list_items in list_of_lists]


def get_link(url: str, link_text: Union[str, int, float]) -> str:
    """Create link (e.g. to a commit)."""

    return f'<a target="_parent" href="{url}">{link_text}</a>'


def build_regex_string(string: str) -> str:
    """Escape some characters and replace * and ? wildcard characters with the
    python regex equivalents."""
    string.replace(".", r"\.")
    string.replace("*", ".+")
    string.replace("?", ".")
    return string


def replace_none(input_list: list, replacement='-') -> list:
    """Sanitise list for display purposes."""
    return [item if item is not None else replacement for item in input_list]


def append_list_to_dict(dictionary: Dict[str, list], key: str, value: list):
    """If key exists in dictionary then append value to it, else add a new key with value.

    :param dictionary: A dictionary to add key and value to.
    :param key: Dictionary key.
    :param value: A value to append to the dictionary list.
    """
    if key in dictionary:
        dictionary[key].extend(value)
    else:
        dictionary[key] = value



from robota_core.config_readers import process_yaml


class TestProcessYaml:
    @staticmethod
    def test_root_dict_substitution():
        """The key to be substituted is in the top level of the dictionary."""
        initial = {"name": "Peter", "greeting": "Hello ${name}"}
        expected = {"name": "Peter", "greeting": "Hello Peter"}

        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_root_dict_substitution_with_int():
        """The key to be substituted is an integer in the top level of the dictionary."""
        initial = {"name": 7, "greeting": "Hello ${name}"}
        expected = {"name": 7, "greeting": "Hello 7"}

        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_double_root_dict_substitution():
        """The key to be substituted occurs twice in the top level of the dictionary."""
        initial = {"name": "Peter", "greeting": "Hello ${name} ${name}"}
        expected = {"name": "Peter", "greeting": "Hello Peter Peter"}

        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_different_double_root_dict_substitution():
        """Two different keys are substituted in the top level of the dictionary."""
        initial = {"name1": "Peter", "name2": "Fred", "greeting": "Hello ${name1} ${name2}"}
        expected = {"name1": "Peter", "name2": "Fred", "greeting": "Hello Peter Fred"}

        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_second_level_dict_substitution():
        """The key to be substituted occurs nested inside another top level dictionary."""
        initial = {"name": "Peter", "greetings": {"arrival": "Hello ${name}",
                                                  "departure": "Goodbye ${name}"}}
        expected = {"name": "Peter", "greetings": {"arrival": "Hello Peter",
                                                   "departure": "Goodbye Peter"}}
        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_second_level_list_substitution():
        """The key to be substituted is in a list inside a top level dictionary."""
        initial = {"name": "Peter", "greeting": ["Hello ${name}", 7]}
        expected = {"name": "Peter", "greeting": ["Hello Peter", 7]}

        actual = process_yaml(initial)
        assert actual == expected

    @staticmethod
    def test_only_top_level_keys():
        """Key substitution only works with top level keys."""
        initial = {"name": {"sub-key": "Ahoy!"}, "greeting": "Hello ${sub-key}"}

        actual = process_yaml(initial)
        assert actual == initial

    @staticmethod
    def test_ignore_list():
        """YAML files can be returned as lists if there are no root keys. The function should
        ignore lists as an input."""
        initial = ["Some data", "more ${data}"]

        actual = process_yaml(initial)
        assert actual == initial



""""Tests for gitlab_tools.py"""
from datetime import datetime, timezone, timedelta

import gitlab
import pytest

from robota_core.gitlab_tools import GitlabServer
from robota_core.repository import Event
from robota_core import commit


class TestOpenGitlab:
    """Test the open gitlab function"""
    bad_url = "http://www.google.com/"
    bad_token = "aaa"

    def test_bad_url(self):
        """A bad URL should raise a different error."""
        with pytest.raises(gitlab.exceptions.GitlabGetError):
            _ = GitlabServer(self.bad_url, self.bad_token)


class TestGetTags:
    @staticmethod
    def test_added_tag_after_deadline():
        # Test for the case that a branch has been added since deadline and so should be removed.
        tags = [commit.Tag({"name": "master", "commit_id": "111"}, "dict"),
                commit.Tag({"name": "develop", "commit_id": "222"}, "dict"),
                commit.Tag({"name": "feature", "commit_id": "333"}, "dict")]

        events = [Event({"date": "2020-01-02T00:00:00.000Z", "type": "pushed new",
                         "push_data": {"ref_type": "tag", "ref_name": "feature",
                                       "commit_id": "333", "commit_count": "1"}})]

        tags = commit.get_tags_at_date(datetime(2020, 1, 1, tzinfo=timezone(timedelta(hours=0))), tags, events)
        assert isinstance(tags, list)
        assert len(tags) == 2
        assert tags[0].name == "master"
        assert tags[1].name == "develop"

    @staticmethod
    def test_deleted_tag_after_deadline():
        # Test for the case that a branch has been deleted since deadline and so should be added.
        tags = [commit.Tag({"name": "master", "commit_id": "111"}, "dict"),
                commit.Tag({"name": "develop", "commit_id": "222"}, "dict")]

        events = [Event({"date": "2020-01-02T00:00:00.000Z", "type": "deleted",
                         "push_data": {"ref_type": "tag", "ref_name": "feature",
                                       "commit_id": "333", "commit_count": "1"}})]

        tags = commit.get_tags_at_date(datetime(2020, 1, 1, tzinfo=timezone(timedelta(hours=0))), tags, events)
        assert isinstance(tags, list)
        assert len(tags) == 3
        assert tags[0].name == "master"
        assert tags[1].name == "develop"
        assert tags[2].name == "feature"

    @staticmethod
    def test_tag_changed_after_deadline():
        # Test for the case that a branch has been added and deleted since deadline
        # and so should not be added.

        tags = [commit.Tag({"name": "master", "commit_id": "111"}, "dict"),
                commit.Tag({"name": "develop", "commit_id": "222"}, "dict")]

        # Events come from gitlab most recent first.
        events = [Event({"date": "2020-01-02T00:01:00.000Z", "type": "deleted",
                         "push_data": {"ref_type": "tag", "ref_name": "feature",
                                       "commit_id": "333", "commit_count": "1"}}),
                  Event({"date": "2020-01-02T00:00:00.000Z", "type": "pushed new",
                         "push_data": {"ref_type": "tag", "ref_name": "feature",
                                       "commit_id": "333", "commit_count": "1"}})
                  ]

        tags = commit.get_tags_at_date(datetime(2020, 1, 1, tzinfo=timezone(timedelta(hours=0))), tags, events)
        assert isinstance(tags, list)
        assert len(tags) == 2
        assert tags[0].name == "master"
        assert tags[1].name == "develop"



from datetime import datetime
from typing import Union, List

from robota_core.remote_provider import RemoteProvider, new_remote_provider
from robota_core.repository import new_repository, Repository
from robota_core.issue import new_issue_server, IssueServer
from robota_core.ci import new_ci_server, CIServer


class DataServer:
    """A container for data sources."""

    def __init__(self, robota_config: dict, start: datetime, end: datetime):
        # Set up connection to student repo
        self.repository: Union[Repository, None] = new_repository(robota_config)
        self.remote_provider: Union[RemoteProvider, None] = new_remote_provider(robota_config)
        self.issue_server: Union[IssueServer, None] = new_issue_server(robota_config)
        self.ci_server: Union[CIServer, None] = new_ci_server(robota_config)

        self.start = start
        self.end = end
        self._valid_sources: Union[None, List[str]] = None

    def get_valid_sources(self) -> List[str]:
        """Check whether each of the data sources has been successfully initialised."""
        if self._valid_sources is not None:
            return self._valid_sources

        # Populate the valid sources list
        self._valid_sources = []
        if self.repository:
            self._valid_sources.append("repository")
        if self.remote_provider:
            self._valid_sources.append("remote_provider")
        if self.issue_server:
            self._valid_sources.append("issues")
        if self.ci_server:
            self._valid_sources.append("ci")
        return self._valid_sources



import datetime
from typing import List, Union

import gitlab
import github.PullRequest
import github.PullRequestComment
import github.IssueComment

from robota_core.string_processing import markdownify, clean, string_to_datetime, get_link


class MergeRequest:
    """A Merge Request"""
    def __init__(self, merge_request, source: str):
        self.number: Union[int, None] = None
        self.source_branch = None
        self.target_branch = None
        self.author = None
        self.url: str = ""
        self.comments = None
        self.state = None

        if source == "gitlab":
            self._merge_request_from_gitlab(merge_request)
        elif source == "github":
            self._merge_request_from_github(merge_request)
        else:
            raise TypeError("Merge request type not recognised.")

        self.link = get_link(self.url, self.number)

    def _merge_request_from_github(self, merge_request: github.PullRequest.PullRequest):
        self.number = merge_request.number
        self.source_branch = merge_request.head.ref
        self.target_branch = merge_request.base.ref
        self.author = merge_request.user.name
        self.url = merge_request.html_url
        comments = list(merge_request.get_issue_comments())
        comments.extend(merge_request.get_comments())
        self.comments = [MergeRequestComment(comment, "github") for comment in comments]
        self.state = merge_request.state

    def _merge_request_from_gitlab(self, gl_merge_request: gitlab.v4.objects.ProjectMergeRequest):
        """Convert a GitLab merge request into a RoboTA merge request"""
        self.number = gl_merge_request.attributes['iid']
        self.source_branch = gl_merge_request.attributes["source_branch"]
        self.target_branch = gl_merge_request.attributes["target_branch"]
        self.author = gl_merge_request.attributes['author']
        self.url = gl_merge_request.attributes['web_url']
        self.comments = [MergeRequestComment(note, "gitlab") for
                         note in gl_merge_request.notes.list(get_all=True)]
        self.state = gl_merge_request.attributes['state']


class MergeRequestCache:
    """A cache of MergeRequest objects from a specific date range."""
    def __init__(self, start: datetime.datetime, end: datetime.datetime,
                 merge_requests: List[MergeRequest]):
        self.start = start
        self.end = end
        self.merge_requests = merge_requests

    def __iter__(self):
        yield from self.merge_requests

    def add_merge_request(self, merge_request: MergeRequest):
        """Add a MergeRequest to a MergeRequestCache."""
        self.merge_requests.append(merge_request)


class MergeRequestComment:
    """Comments on a merge request"""
    def __init__(self, comment, source: str):
        self.body = None
        self.author = None
        self.created_at = None

        if source == "gitlab":
            self._comment_from_gitlab_merge_request(comment)
        elif source == "github":
            self._comment_from_github_merge_request(comment)
        else:
            raise TypeError("Merge request type not recognised.")

    def _comment_from_github_merge_request(self, gh_mr_note: Union[
          github.PullRequestComment.PullRequestComment, github.IssueComment.IssueComment]):
        self.body = markdownify(clean(gh_mr_note.body))
        self.author = gh_mr_note.user.name
        self.created_at = gh_mr_note.created_at

    def _comment_from_gitlab_merge_request(self,
                                           gl_mr_note: gitlab.v4.objects.ProjectMergeRequestNote):
        self.body = markdownify(clean(gl_mr_note.attributes['body']))
        self.author = gl_mr_note.attributes['author']
        self.created_at = string_to_datetime(gl_mr_note.attributes['created_at'])



"""Objects and for describing and processing Git Issues."""
from abc import abstractmethod
import datetime
from operator import attrgetter
from typing import List, Union, Tuple
import re

import github.Issue
import github.IssueComment
from gitlab.v4.objects import ProjectIssueNote, ProjectIssue
from loguru import logger

from robota_core import gitlab_tools, config_readers
from robota_core.github_tools import GithubServer
from robota_core.string_processing import string_to_datetime, get_link, clean


class Issue:
    """An Issue

    :ivar created_at: (datetime) The time at which the issue was created.
    :ivar assignee: (string) The person to whom the issue was assigned.
    :ivar closed_at: (datetime) The time at which the issue was closed.
    :ivar closed_by: (string) The person who closed the issue.
    :ivar time_stats: (dict) Estimates and reported time taken to work on the issue.
    :ivar due_date: (datetime) The time at which issue is due to be completed.
    :ivar title: (string) The title of the issue.
    :ivar comments: (List[Comment]) A list of Comments associated with the Issue.
    :ivar state: (string) Whether the issue is open or closed.
    :ivar milestone: (string) Any milestone the issue is associated with.
    :ivar url: (string) A link to the Issue on GitLab.

    """
    def __init__(self, issue, issue_source: str, get_comments=True):
        self.created_at = None
        self.assignee = None
        self.closed_at = None
        self.closed_by = None
        self.time_stats = None
        self.due_date = None
        self.title = ""
        self.comments: List[IssueComment] = []
        self.state = None
        self.milestone = None
        self.url = ""
        self.number = None

        if issue_source == "gitlab":
            self._issue_from_gitlab(issue, get_comments)
        elif issue_source == "github":
            self._issue_from_github(issue, get_comments)
        elif issue_source == "test data":
            self._issue_from_test_data(issue)
        else:
            raise TypeError(f"Unknown issue type: '{issue_source}'")

        self.link = get_link(self.url, self.title)

    def __eq__(self, other_issue: Union[None, "Issue"]) -> bool:
        if other_issue is None:
            return False
        elif self.created_at == other_issue.created_at and self.title == other_issue.title:
            return True
        else:
            return False

    def __repr__(self) -> str:
        return f"Issue: {self.title}"

    def _issue_from_github(self, github_issue: github.Issue.Issue, get_comments: bool):
        self.created_at = github_issue.created_at
        if github_issue.assignee:
            self.assignee = github_issue.assignee.name
        else:
            self.assignee = None
        if github_issue.closed_by:
            self.closed_by = github_issue.closed_by.name
        else:
            self.closed_by = None
        self.title = github_issue.title
        self.state = github_issue.state
        if github_issue.milestone:
            self.milestone = {"title": github_issue.milestone.title,
                              'web_url': github_issue.milestone.url}
        else:
            self.milestone = None
        self.url = github_issue.html_url
        self.number = github_issue.number
        if get_comments:
            comments = github_issue.get_comments()
            for comment in comments:
                self.comments.append(IssueComment(comment, "github"))

    def _issue_from_gitlab(self, gitlab_issue: ProjectIssue, get_comments: bool):
        """Convert a GitLabIssue to a RoboTA issue."""
        self.created_at = string_to_datetime(gitlab_issue.attributes["created_at"])
        self.assignee = gitlab_issue.attributes["assignee"]
        self.closed_at = string_to_datetime(gitlab_issue.attributes["closed_at"])
        self.closed_by = gitlab_issue.attributes["closed_by"]
        self.time_stats = gitlab_issue.attributes["time_stats"]
        due_date_as_datetime = gitlab_issue.attributes["due_date"] + "T12:00:00.000+00:00" \
                               if gitlab_issue.attributes["due_date"] is not None \
                               else None
        self.due_date = string_to_datetime(due_date_as_datetime)
        self.title = gitlab_issue.attributes["title"]
        if gitlab_issue.state == "opened":
            self.state = "open"
        else:
            self.state = gitlab_issue.state
        gitlab_milestone = gitlab_issue.attributes["milestone"]
        if gitlab_milestone:
            self.milestone = {"title": gitlab_milestone['title'],
                              'web_url': gitlab_milestone['web_url']}

        self.url = gitlab_issue.attributes["web_url"]
        self.number = gitlab_issue.attributes["iid"]
        if get_comments:
            all_notes = gitlab_issue.notes.list(all=True)
            for note in all_notes:
                self.comments.append(IssueComment(note, "gitlab"))

            # Convert the issue state events into comments (since GitLab now handles these events separately
            # and does not create a comment on the issue when its state changes.
            all_state_events = gitlab_issue.resourcestateevents.list(all=True)
            for state_change in all_state_events:
                state_change_occurred_at = string_to_datetime(state_change.created_at)
                note = (state_change.state, state_change_occurred_at, state_change_occurred_at, True, state_change.user["username"])
                self.comments.append(IssueComment(note, "test data"))

        # Returns comments in descending order of creation date (oldest first)
        self.comments.sort(key=attrgetter("created_at"))

    def _issue_from_test_data(self, issue_data):
        (number, title) = issue_data
        self.number = number
        self.title = title

    def get_assignee(self) -> Union[str, None]:
        """ Return name of issue assignee

        :return If issue has an assignee, returns their name else returns None.
        """
        if self.assignee:
            return self.assignee['name']
        return None

    def get_assignment_date(self) -> Union[datetime.datetime, None]:
        """Get assignment date for an issue.
        First checks comments for assignment date and if none is found, returns the issue creation
        date. If there is more than one assignment date, this method will always return the most
        recent.

        :return: The date at which the issue was assigned.
        """
        if not self.assignee:
            return None

        # Looking for most recent comment first so reverse comment list.
        for comment in reversed(self.comments):
            if comment.text.startswith('assigned to'):
                return comment.created_at
        return self.created_at

    def get_time_estimate_date(self) -> Union[datetime.datetime, None]:
        """Gets the date a time estimate was added to an issue. This only works for issues
        made after 05/02/19 as this was a feature added in Gitlab 11.4.

        :return: Date of the first time estimate, None if no time estimate was found.
        """
        # Comments are stored oldest first.
        for comment in reversed(self.comments):
            if comment.text.startswith('changed time estimate to'):
                return comment.created_at
        return None

    def get_time_estimate(self) -> datetime.timedelta:
        """Gets estimate of time it will take to close issue."""
        time_estimate = self.time_stats['time_estimate']
        return datetime.timedelta(seconds=time_estimate)

    def get_comment_timestamp(self, key_phrase: str,
                              earliest=False) -> Union[datetime.datetime, None]:
        """Search for a phrase in the comments of an issue
        If the phrase exists, return creation time of the comment.

        :param key_phrase: a phrase to search for in a comment on the issue.
        :param earliest: If True, return the earliest comment matching key_phrase, else return
          most recent comment matching key_phrase.
        :return: If phrase is present in a comment, return the the time of the comment,
          else return None
        """
        # Since we can't guarantee that all Git hosting sites will return comments in the same order
        # we specifically search for the one we want
        comments = self.comments

        matching_comments = [c.created_at for c in comments if key_phrase in c.text]
        if matching_comments:
            matching_comments.sort(reverse=not earliest)
            return matching_comments[0]
        return None

    def get_recorded_team_member(self, key_phrase: str) -> Union[None, List[str]]:
        """Report whether a team member has been recorded using a key phrase for issue.
        Key phrase should appear at the start of a comment to indicate assignment of sub-team
        member, code reviewer (etc).

        :param key_phrase: Phrase to search for
        :return team_member_recorded: Str
        """

        # Strings we're searching for are:
        # - key_phrase @username
        # - key_phrase https://gitlab.cs.man.ac.uk/username
        # - key_phrase https://gitlab.cs.man.ac.uk/user.name
        # - key_phrase https://gitlab.cs.man.ac.uk/user-name
        # Also permit the team member to be quoted or in angle brackets
        # Also permit the url to be in square brackets as this is markdown for a link
        regex = r"\s*(<|\"|\'|\[)*(@|https:\/\/gitlab\.cs\.man\.ac\.uk\/)(\w+[-\.]?\w*)(>|\"|\'|\])*"
        regex = key_phrase + regex
        recorded_team_member = []
        for comment in self.comments:
            match = re.findall(regex, comment.text)
            if match:
                for match_contents in match:
                    recorded_team_member.append(match_contents[2])

        if recorded_team_member:
            return recorded_team_member
        return None

    def is_assignee_contributing(self, team) -> Union[bool, str]:
        """Determine whether the Student assigned to work on an Issue is contributing to the
        exercise."""
        if self.assignee is None:
            return "No issue assignee."
        else:
            assigned_student = team.get_student_by_name(self.assignee["name"])
            if assigned_student is None:
                # This will happen if a student leaves the team after the exercise.
                return "Assignee is not a team member"
            else:
                return assigned_student.is_contributing

    def get_status(self, deadline: datetime.datetime):
        """Get current status of issue if deadline hasn't passed,
        otherwise get last status of issue before the deadline,
        and save in the issue.state attribute so that it is only calculated once.

        :param deadline:
        :return:
        """
        if datetime.datetime.now(datetime.timezone.utc) < deadline:
            return self.state
        else:
            for comment in self.comments:
                # Has the issue status changed since the deadline?
                if comment.system and comment.created_at < deadline:
                    if comment.text.startswith('closed'):
                        self.state = 'closed'
                        break
                    elif comment.text == 'reopened':
                        self.state = 'open'
                        break
            else:
                # No status change before the deadline
                self.state = 'open'

            return self.state


class IssueCache:
    """A cache of Issue objects from a specific date range."""
    def __init__(self, start: datetime.datetime = None, end: datetime.datetime = None,
                 get_comments=True, milestone=None):
        self.start = start
        self.end = end
        self.get_comments = get_comments
        self.issues: List[Issue] = []
        self.milestone = milestone

    def __iter__(self):
        yield from self.issues

    def add_issue(self, issue: Issue):
        """Add an Issue to an IssueCache."""
        self.issues.append(issue)


class IssueServer:
    """An IssueServer is a service from which Issues are extracted."""
    def __init__(self):
        self._stored_issues: List[IssueCache] = []

    def get_issues(self, start: datetime.datetime = datetime.datetime.fromtimestamp(1, datetime.timezone.utc),
                   end: datetime.datetime = datetime.datetime.now(datetime.timezone.utc),
                   get_comments: bool = True) -> List[Issue]:
        """Get issues from the issue provider between the start date and end date."""
        cached_issues = self._get_cached_issues(start, end)
        if cached_issues:
            return cached_issues.issues

        new_issues = self._fetch_issues(start, end, get_comments)
        cached_issues = IssueCache(start, end, get_comments)
        for issue in new_issues:
            cached_issues.add_issue(issue)
        self._stored_issues.append(cached_issues)
        return new_issues

    def get_issues_by_milestone(self, milestone_name: str) -> Union[List[Issue], None]:
        """Get a list of issues associated with a milestone."""
        for issue_cache in self._stored_issues:
            if issue_cache.milestone == milestone_name:
                return issue_cache.issues

        new_issues = self._fetch_issues_by_milestone(milestone_name)
        new_cache = IssueCache(milestone=milestone_name)
        for issue in new_issues:
            new_cache.add_issue(issue)
        self._stored_issues.append(new_cache)
        return new_issues

    @abstractmethod
    def _fetch_issues(self, start: datetime.datetime, end: datetime.datetime,
                      get_comments: bool) -> List[Issue]:
        """Get issues from the issue provider between the start date and end date."""
        raise NotImplementedError("Not implemented in base class.")

    @abstractmethod
    def _fetch_issues_by_milestone(self, milestone_name: str) -> List[Issue]:
        """Get issues associated with the given milestone from the issue provider."""
        raise NotImplementedError("Not implemented in base class.")

    def _get_cached_issues(self, start: datetime.datetime,
                           end: datetime.datetime) -> Union[IssueCache, None]:
        """Check whether issues with the specified start and end date are already stored."""
        for cache in self._stored_issues:
            if cache.start and cache.end:
                if cache.start == start and cache.end == end:
                    return cache
        else:
            return None


class GitLabIssueServer(IssueServer):
    """An IssueServer with GitLab as the server."""

    def __init__(self, issue_source: dict):
        super().__init__()
        if "token" in issue_source:
            token = issue_source["token"]
        else:
            token = None
        gitlab_server = gitlab_tools.GitlabServer(issue_source["url"], token)
        self.project = gitlab_server.open_gitlab_project(issue_source["project"])

    def _fetch_issues(self, start: datetime.datetime, end: datetime.datetime,
                      get_comments=True) -> List[Issue]:
        """Function to return issues falling withing a certain time window.

        :param start: The start of the time window for included issues
        :param end: The end of the time window for included issues.
        :param get_comments: Whether or not to download issue comments from the server.
          This may take some time if there are a large number of issues so should be disabled
          if the comments are not needed.
        :return: A list of Issue objects.
        """
        request_parameters = {}
        if start is not None:
            request_parameters['created_after'] = start.isoformat()
        if end is not None:
            request_parameters['created_before'] = end.isoformat()

        gitlab_issues = self.project.issues.list(all=True,
                                                 query_parameters=request_parameters)

        return [Issue(gitlab_issue, "gitlab", get_comments) for gitlab_issue in gitlab_issues]

    def _fetch_issues_by_milestone(self, milestone_name: str) -> List[Issue]:
        """Get all gitlab issues associated with a particular milestone.

        :param milestone_name: The name of the milestone to find.
        """
        project_milestones = self.project.milestones.list()

        for milestone in project_milestones:
            if milestone.attributes["title"] == milestone_name:
                milestone_issues = list(milestone.issues())
                return [Issue(issue, "gitlab") for issue in milestone_issues]
        # If the milestone exists but there are no issues associated with it.
        return []


class GitHubIssueServer(IssueServer):
    def __init__(self, issue_server_source: dict):
        super().__init__()
        server = GithubServer(issue_server_source)
        self.repo = server.open_github_repo(issue_server_source["project"])

    def _fetch_issues(self, start: datetime.datetime, end: datetime.datetime,
                      get_comments: bool) -> List[Issue]:
        # TODO: This method does not check issue [opening] end date
        issues = self.repo.get_issues(state="all", since=start)
        return [Issue(issue, "github") for issue in issues if not issue.pull_request]

    def _fetch_issues_by_milestone(self, milestone_name: str) -> List[Issue]:
        milestones = self.repo.get_milestones()
        for milestone in milestones:
            if milestone.title == milestone_name:
                issues = self.repo.get_issues(milestone=milestone, state="all")
                return [Issue(issue, "github") for issue in issues]
        # If milestone not found
        return []


class IssueComment:
    """A comment is a textual field attached to an Issue

    :ivar text: (string) The content of the comment message.
    :ivar created_at: (datetime) The time a comment was made.
    :ivar updated_at: (datetime) The most recent time the content of a comment was updated.
    """
    def __init__(self, comment, source: str):
        self.text = None
        self.created_at = None
        self.updated_at = None
        self.system = None
        self.author = None

        if source == "gitlab":
            self._comment_from_gitlab(comment)
        elif source == "github":
            self._comment_from_github(comment)
        elif source == "test data":
            self._comment_from_test_data(comment)
        else:
            raise TypeError(f"Unknown commit comment source: '{source}'.")

    def _comment_from_gitlab(self, comment: ProjectIssueNote):
        """Populate an instance of a comment from a GitLab note."""
        self.text = clean(comment.attributes["body"])
        self.created_at = string_to_datetime(comment.attributes["created_at"])
        self.updated_at = string_to_datetime(comment.attributes["updated_at"])
        self.system = comment.attributes["system"]
        self.author = comment.attributes["author"]["username"]

    def _comment_from_github(self, comment: github.IssueComment):
        self.text = comment.body
        self.created_at = comment.created_at
        self.updated_at = comment.updated_at
        self.author = comment.user["login"]

    def _comment_from_test_data(self, comment: Tuple[str, datetime.datetime, datetime.datetime, str, str]):
        (text, created_at, updated_at, system, author_name) = comment
        self.text = text
        self.created_at = created_at
        self.updated_at = updated_at
        self.system = system
        self.author = author_name


def get_issue_by_title(issues: List[Issue], title: str) -> Union[Issue, None]:
    """If issue with 'title' exists in 'issues', return the issue, else return None.

    :param issues: A list of Issue objects.
    :param title: An issue title
    :returns: Issue with title == title, else None.
    """
    for issue in issues:
        if issue.title == title:
            return issue
    return None


def new_issue_server(robota_config: dict) -> Union[None, IssueServer]:
    """A factory method for IssueServers."""
    issue_server_source = config_readers.get_data_source_info(robota_config, 'issues')
    if not issue_server_source:
        return None
    server_type = issue_server_source["type"]
    logger.debug(f"Initialising {server_type} issue server.")

    if server_type == 'gitlab':
        return GitLabIssueServer(issue_server_source)
    if server_type == 'github':
        return GitHubIssueServer(issue_server_source)
    else:
        raise TypeError(f"Unknown issue server type {server_type}.")



# Unit and simple integration tests for functionality concerning issue handling
from datetime import datetime, timedelta
from unittest import TestCase

from robota_core.issue import Issue, IssueComment


class TestIssueHandling(TestCase):
    @staticmethod
    def test_fetch_earliest_matching_comment():
        text_to_match = "changed due date to"
        first_date = datetime(2020, 1, 1, hour=12)
        one_day = timedelta(days=1)
        two_days = timedelta(days=2)
        update_date = datetime(2020, 5, 5)
        test_issue = Issue((1, 'GUI bug'), "test data")
        author1 = "jemima.puddleduck"
        author2 = "jfisher572"
        # A list of comments in non-chronological order
        test_issue.comments = [
            IssueComment(("removed milestone", first_date + one_day, update_date, True, author1), "test data"),
            IssueComment((text_to_match, first_date + two_days, update_date, True, author1), "test data"),
            IssueComment(("Tests completed for this issue", first_date - one_day, update_date, False, author2), "test data"),
            IssueComment((text_to_match, first_date, update_date, True, author2), "test data"),
            IssueComment(("changed time estimate to", first_date - two_days, update_date, True, author2), "test data")
        ]

        assert test_issue.get_comment_timestamp(text_to_match, earliest=True) == first_date
        assert test_issue.get_comment_timestamp(text_to_match, earliest=False) == first_date + two_days

    def test_identify_sub_team_membership_in_issue_comments(self):
        key_phrase = 'Sub team member:'
        # Set up an issue with comments covering all the key cases.
        # We don't care about the order of these comments or when they were made, so use a dummy date
        test_issue = Issue((1, 'GUI bug'), "test data")
        dummy_date = None
        dummy_author = "anne.author"
        test_issue.comments = [
            # An irrelevant comment
            IssueComment(("removed milestone", dummy_date, dummy_date, True, dummy_author), "test data"),
            # An irrevelvant comment containing a username
            IssueComment(("Code review by: @a12345bc", dummy_date, dummy_date, True, dummy_author), "test data"),

            # A sub-team assignment of the form: key_phrase @username
            IssueComment((key_phrase + " @d23456ef", dummy_date, dummy_date, True, dummy_author), "test data"),

            # A sub-team assignment of the form: key_phrase https://gitlab.cs.man.ac.uk/username
            IssueComment((key_phrase + " https://gitlab.cs.man.ac.uk/g34567hi", dummy_date, dummy_date, True, dummy_author), "test data"),
            # A sub-team assignment of the form: key_phrase https://gitlab.cs.man.ac.uk/user.name
            IssueComment((key_phrase + " https://gitlab.cs.man.ac.uk/some.one", dummy_date, dummy_date, True, dummy_author), "test data"),
            # A sub-team assignment of the form: key_phrase https://gitlab.cs.man.ac.uk/user-name
            IssueComment((key_phrase + " https://gitlab.cs.man.ac.uk/person-two", dummy_date, dummy_date, True, dummy_author), "test data"),

            # A sub-team assignment of the form: key_phrase [https://gitlab.cs.man.ac.uk/username]
            IssueComment((key_phrase + " [https://gitlab.cs.man.ac.uk/j45678kl]", dummy_date, None, True, dummy_author), "test data"),
            # A sub-team assignment of the form: key_phrase <https://gitlab.cs.man.ac.uk/user.name>
            IssueComment((key_phrase + " <https://gitlab.cs.man.ac.uk/m56789no>", dummy_date, None, True, dummy_author), "test data")
        ]

        actualTeamMembers = test_issue.get_recorded_team_member(key_phrase)
        expectedTeamMembers = ["d23456ef", "g34567hi", "some.one", "person-two", "j45678kl", "m56789no"]
        self.assertCountEqual(actualTeamMembers, expectedTeamMembers)



"""General methods for interfacing with GitLab via the python-Gitlab library."""
import sys

from loguru import logger
import urllib.request
from typing import List

import gitlab.v4.objects


class GitlabGroup:
    """ A group is distinct from a project, a group may contain many projects.
    Projects contained in a group inherit the members of the containing project.
    """
    def __init__(self, gitlab_connection: gitlab.Gitlab, group_name: str):
        self.group_name = group_name
        self.group = gitlab_connection.groups.get(group_name)

    def get_group_members(self) -> List[str]:
        """ Get a list of members in a group.

        :return member_list: Names of members of group.
        """
        member_list = []
        for member in self.group.members.list():
            member_list.append(member.attributes['name'])

        return member_list


class GitlabServer:
    """A connection to the Gitlab server. Contains methods for interfacing with the API. This is
    held distinct from the Repository object because it can also be used to interface with
    an Issue server."""
    def __init__(self, url: str, token: str):
        """ Initialise the connection to the server, getting credentials from the credentials file.

        :param url: url of GitLab server
        :param token: Authentication token for gitlab server.
        """
        self.url = url
        self.token = token

        self.gitlab_connection: gitlab.Gitlab = self._open_gitlab_connection()

    def _open_gitlab_connection(self) -> gitlab.Gitlab:
        """Open a connection to the GitLab server using authentication token."""
        if not self.token:
            logger.error("Must provide an authentication token in robota config to use the "
                         "gitlab API.")
            raise KeyError()
        server = gitlab.Gitlab(self.url, private_token=self.token)
        try:
            server.auth()
        except gitlab.exceptions.GitlabAuthenticationError:
            logger.error("Incorrect authentication token provided. Unable to connect to GitLab.")
            # Exit to prevent really long gitlab stack trace.
            sys.exit(1)
        except gitlab.exceptions.GitlabHttpError:
            logger.error(f"Unable to find gitlab server '{self.url}.")
            sys.exit(1)

        logger.info(f"Logged in to gitlab: '{server.url}' as {server.user.attributes['name']}")

        return server

    def open_gitlab_project(self, project_path: str) -> gitlab.v4.objects.Project:
        """Open a GitLab project.

        :param project_path: The path of the project to open. Includes namespace.
        :return: A GitLab project object.
        """
        if "/" not in project_path:
            raise gitlab.exceptions.GitlabGetError("Must provide namespace "
                                                   "when opening gitlab project.")
        try:
            url_encoded_path = urllib.request.pathname2url(project_path)
            project = self.gitlab_connection.projects.get(url_encoded_path)
        except gitlab.exceptions.GitlabGetError as error_type:
            logger.error(f"Unable to find project: {project_path}. It either does not exist or "
                         f"the current user {self.gitlab_connection.user.attributes['name']} does "
                         f"not have access to this project.")
            sys.exit(1)
        logger.info(f"Connected to gitlab project {project.attributes['path_with_namespace']}")
        return project

    def open_gitlab_group(self, group_name: str) -> GitlabGroup:
        return GitlabGroup(self.gitlab_connection, group_name)



"""Test for functions in the robota_tools.py file"""
import pathlib
import pytest

import robota_core.config_readers as config_readers

TEST_DIR = pathlib.Path(__file__).parent
ROOT_DIR = TEST_DIR.parent.parent.parent
TEST_FOLDER = TEST_DIR / pathlib.Path("sample_files")
REPO_URL = "https://gitlab.cs.man.ac.uk/institute-of-coding-team/comp23412-robota-config.git"


class TestGetRobotaConfig:
    """Test the functions in the robota_core.robota_tools.GetRobotaConfig class"""
    @staticmethod
    @pytest.mark.config_path
    def test_get_config_bad():
        """Integration test to check exception is thrown when an unknown type is loaded."""
        dummy_config_path = "/dev/null"
        with pytest.raises(config_readers.RobotaConfigLoadError):
            _ = config_readers.get_config(["config.yaml"], {"type": "aaaaaa",
                                                            "directory": dummy_config_path})


class TestReadConfigFile:
    """Test the functions in robota_core.robota_tools._ReadConfigFile"""
    @staticmethod
    def test_yaml_file_read():
        """Test that a valid yaml file can be read"""
        test_config_path = TEST_FOLDER / pathlib.Path("config.yaml")
        config = config_readers.read_yaml_file(test_config_path)
        assert isinstance(config, dict)
        assert config["Name"] == "Fred"
        assert config["Year"] == 2019

    @staticmethod
    def test_config_parser_valid():
        """Check the parse_config method with a valid file type"""
        test_config_path = TEST_FOLDER / pathlib.Path("config.yaml")
        config = config_readers.parse_config(test_config_path)

        assert config["Name"] == "Fred"
        assert config["Year"] == 2019

    @staticmethod
    def test_config_parser_invalid():
        """Check an exception is raised if an invalid file type is passed."""
        with pytest.raises(config_readers.RobotaConfigParseError):
            bad_config_path = TEST_FOLDER / pathlib.Path("another_config.txt")
            _ = config_readers.parse_config(bad_config_path)



"""A module for visualising the commits in git repositories and their relations."""
import datetime
import subprocess
from typing import List, TextIO, Tuple

from robota_core.config_readers import get_robota_config
from robota_core.repository import new_repository


def catch_empty_graph(nodes, all_commits, commit_parents):
    """If there are no branches and no tags, print the commits starting from the oldest."""
    if not nodes:
        parent_id = all_commits[-1]
        commit_list = [parent_id]
        parents = True
        while parents:
            if parent_id in all_commits:
                parent_index = all_commits.index(parent_id)
                parents = commit_parents[parent_index]
                parent_id = parents[0]
                commit_list.append(parent_id)
            else:
                parents = False
        nodes.append(list(reversed(commit_list)))
    return nodes


def identify_merge_commit_parents(commit_parents: List[List[str]]) -> List[List[str]]:
    """From the list of all commits and commit parents, find the merge commits and then return
    a list of parents of these merge commits.
    :param commit_parents: a list of commit parents corresponding to all commits.
    :returns: a list of merge commit parents, two parents for each merge commit.
    """
    merge_commit_parents = []
    for parents in commit_parents:
        if len(parents) > 1:
            merge_commit_parents.append(parents)
    return merge_commit_parents


def get_branch_commits(all_commits, commit_parents, merge_commit_parents):
    """After identifying the parents of merge commits, plot out each branch by following
    the commits back in time, taking the first parent if there is a choice of two commits.
    This is roughly equivalent to the git command:
    git log --reverse --first-parent --pretty=format:"%h" commit_id
    where commit_id is the child commit. of the two parents."""
    nodes = []
    for parent_pair in merge_commit_parents:
        for parent_id in parent_pair:
            child_index = commit_parents.index(parent_pair)
            child_id = all_commits[child_index]
            commit_list = [child_id, parent_id]
            parents = True
            while parents:
                if parent_id in all_commits:
                    parent_index = all_commits.index(parent_id)
                    parents = commit_parents[parent_index]
                    parent_id = parents[0]
                    commit_list.append(parent_id)
                else:
                    parents = False
            nodes.append(list(reversed(commit_list)))
    return nodes


def add_unmerged_branches(refs, all_commits, commit_parents, nodes):
    """Unmerged branches are not identified since they do not have a merge commit.
    Unmerged branches can't exist without a ref so we can find them by going through
    all of the refs."""
    flat_parents = [commit_id for sublist in commit_parents for commit_id in sublist]
    for branch_name in refs:
        if refs[branch_name] not in flat_parents:
            branch = [refs[branch_name]]
            if refs[branch_name] in all_commits:
                parent_index = all_commits.index(refs[branch_name])
                parents = commit_parents[parent_index]
                while parents:
                    parent_id = parents[0]
                    branch.append(parent_id)
                    if parent_id in all_commits:
                        parent_index = all_commits.index(parent_id)
                        parents = commit_parents[parent_index]
                    else:
                        parents = False
                nodes.append(list(reversed(branch)))
    return nodes


def output_dot_file(nodes, refs, all_commits):
    """Makes the output file from the collected information."""
    with open("output.dot", 'w') as output_file:
        output_nodes(nodes, output_file)
        output_refs(refs, output_file, all_commits)


def output_nodes(nodes: List[List], output_file: TextIO):
    """Output a representation of the commits.
    :param nodes: A list of commits in each branch of the repository.
    :param output_file: An open file handle to write to.
    """
    output_file.write("strict digraph example {\n")
    for branch_index, branch in enumerate(nodes):
        output_file.write(f'\tnode[group="{branch_index}"];\n\t')
        for commit_index, commit_id in enumerate(branch):
            output_file.write(f'"{commit_id}"')
            if commit_index < (len(branch) - 1):
                output_file.write(" -> ")
            else:
                output_file.write(";\n")
    output_file.write("\n")


def output_refs(refs: dict, output_file: TextIO, all_commits):
    """Output ref labels to decorate the nodes.
    :param refs: A dict of name: commit_id pairs describing repository refs.
    :param output_file: An open file handle to write to.
    """
    for ref_index, ref in enumerate(refs):
        if refs[ref] in all_commits:
            output_file.write(f'\tsubgraph Decorate{ref_index}\n\t{{\n')
            output_file.write('\t\trank = "same";\n')
            output_file.write(f'\t\t"({ref})" [shape = "box", style = "filled", '
                              f'fillcolor = "#ddddff"];\n')
            output_file.write(f'\t\t"({ref})" -> "{refs[ref]}" [weight = 0, arrowtype = "none", '
                              f'dirtype = "none", arrowhead = "none", style = "dotted"];\n\t}}\n')
    output_file.write("}\n")


def render():
    """Render the generated dot file"""
    subprocess.run('dot -Tpng -Gdpi=150 output.dot -o output.png')


def process_commits(all_commits: List[str], commit_parents: List[List[str]], refs: dict):
    """ Process commit data to produce a group of nodes and edges to be plotted by GraphVis.
    :param all_commits: A list of commit ids with most recent first,
    these are given by: git log --all --pretty=format:"%h"
    :param commit_parents: The parents of each commit in all commits.
     Given by: git log --all --pretty=format:"%p"
    :param refs: Given by: git for-each-ref --format="'%(refname:short)': '%(objectname:short),'".
    These are used for labelling but also to catch any branches which are unmerged.
    """
    merge_commit_parents = identify_merge_commit_parents(commit_parents)
    nodes = get_branch_commits(all_commits, commit_parents, merge_commit_parents)
    nodes = add_unmerged_branches(refs, all_commits, commit_parents, nodes)
    nodes = catch_empty_graph(nodes, all_commits, commit_parents)
    return nodes


def get_data_from_gitlab(gitlab_project, start_date: str,
                         end_date: str) -> Tuple[List[str], List[List[str]], dict]:
    """Fetch commit and ref data from GitLab matching the specified group and dates."""

    request_parameters = {'since': datetime.datetime.strptime(start_date, "%d/%m/%y").isoformat(),
                          'until': datetime.datetime.strptime(end_date, "%d/%m/%y").isoformat(),
                          'all': True}
    gitlab_commits = gitlab_project.commits.list(all=True, query_parameters=request_parameters)
    gitlab_branches = gitlab_project.branches.list(all=True)
    all_commits = []
    commit_parents = []
    refs = {}
    # The order of the commit IDs doesnt matter as long as the commit IDs
    # line up with the parent IDs.
    for commit in gitlab_commits:
        all_commits.append(commit.attributes["short_id"])
        parents = []
        for parent in commit.attributes["parent_ids"]:
            parents.append(parent[:8])
        commit_parents.append(parents)
    for branch in gitlab_branches:
        refs[branch.attributes["name"]] = branch.attributes["commit"]["short_id"]
    return all_commits, commit_parents, refs


def main():
    """Main entry point for commit graph plotting."""
    config_path = "../../../robota-marking/robota-config.yaml"
    start_date = "01/10/19"
    end_date = "17/10/19"
    team_name = "S1Team01"

    robota_config = get_robota_config(config_path, {"team_name": team_name})

    gitlab_repo = new_repository(robota_config)

    all_commits, commit_parents, refs = get_data_from_gitlab(gitlab_repo.gitlab_project,
                                                             start_date, end_date)
    events = gitlab_repo.get_events()

    # If any parent is empty then invent a fake parent.
    for index, parents in enumerate(commit_parents):
        if not parents:
            commit_parents[index] = ["None"]

    nodes = process_commits(all_commits, commit_parents, refs)

    output_dot_file(nodes, refs, all_commits)

    render()


if __name__ == '__main__':
    main()



import pytest

from robota_core.commit_visualisation import commit_visualisation


class TestMain:
    def test_no_branch_no_refs(self):
        all_commits = ["a", "b", "c", "d", "e", "f"]
        commit_parents = [["x"], ["a"], ["b"], ["c"], ["d"], ["e"]]
        refs = {}
        commit_visualisation.process_commits(all_commits, commit_parents, refs)
        try:
            commit_visualisation.render()
        except FileNotFoundError:
            pytest.skip("Graphvis executable not found")

    def test_no_branch(self):
        all_commits = ["a", "b", "c", "d", "e", "f"]
        commit_parents = [["x"], ["a"], ["b"], ["c"], ["d"], ["e"]]
        refs = {"master": "f"}

        commit_visualisation.process_commits(all_commits, commit_parents, refs)
        try:
            commit_visualisation.render()
        except FileNotFoundError:
            pytest.skip("Graphvis executable not found")

    def test_one_branch(self):
        all_commits = ["a", "b", "c", "d", "e", "f"]
        commit_parents = [["x"], ["a"], ["b"], ["c"], ["c", "d"], ["e"]]
        refs = {"master": "f", "feature": "d"}

        commit_visualisation.process_commits(all_commits, commit_parents, refs)
        try:
            commit_visualisation.render()
        except FileNotFoundError:
            pytest.skip("Graphvis executable not found")


class TestGetCommitParents:
    def test_no_branch(self):
        all_commits = ["a", "b", "c", "d", "e", "f"]
        commit_parents = [["x"], ["a"], ["b"], ["c"], ["d"], ["e"]]

        merge_commit_parents = commit_visualisation.identify_merge_commit_parents(commit_parents)
        assert merge_commit_parents == []

    def test_one_branch(self):
        all_commits = ["a", "b", "c", "d", "e", "f"]
        commit_parents = [["x"], ["a"], ["b"], ["c"], ["c", "d"], ["e"]]

        merge_commit_parents = commit_visualisation.identify_merge_commit_parents(commit_parents)
        assert merge_commit_parents == [["c", "d"]]


from loguru import logger

import github
import github.Repository

from robota_core import RemoteProviderError


class GithubServer:
    """A connection to GitHub. Contains methods for interfacing with the API."""
    def __init__(self, setup: dict):
        """ Initialise the connection to the server, getting credentials from the credentials file.

        :param setup: dictionary containing GitHub url and authentication token.
        """
        self.url = setup["url"]
        self.server = self._open_gitlab_connection(setup)

    @staticmethod
    def _open_gitlab_connection(setup: dict):
        """Open a connection to the GitLab server using authentication token."""
        token = None
        if "token" in setup:
            token = setup["token"]
        else:
            logger.warning("No auth token provided for Github. Beware that API limits are very "
                            "low for unauthenticated users.")
        return github.Github(token)

    def open_github_repo(self, project_path: str) -> github.Repository.Repository:
        """Open a GitLab repo.

        :param project_path: The path of the project to open. Includes namespace.
        :return: A GitLab project object.
        """
        try:
            repo = self.server.get_repo(project_path)
        except github.UnknownObjectException:
            raise RemoteProviderError(f"Unable to find project: {project_path}. "
                                      f"It either does not exist or the current user does not have "
                                      f"access to this project.")

        logger.info(f"Connected to project {repo.name}")
        return repo



"""A module to collect statistics on student attendance."""
import json
from loguru import logger
import time

import requests

from robota_core import config_readers


class AttendanceError(Exception):
    """An error in collecting attendance data."""


class StudentAttendance:
    """The student attendance class collects data from an external API about student attendance."""

    def __init__(self, robota_config: dict, mock: bool = False):
        """
        :param robota_config: A dictionary of information about data sources read from the robota
          config file.
        :param mock: If True, return mock data instead of getting real data from the data source.
        """
        data_source_name = "attendance"
        attendance_source = config_readers.get_data_source_info(robota_config, data_source_name)
        if not attendance_source:
            raise KeyError(f"Data source '{data_source_name}' not found in robota config.")
        self.data = None
        self.mock = mock
        self._get_course_attendance(attendance_source)
        self.total_sessions = self._get_number_of_sessions()

    def _get_course_attendance(self, attendance_source: dict):
        """Get student attendance using the specified data source.

        :param attendance_source: Information about where to get attendance data from.
        """
        if self.mock:
            logger.warning("Attendance mocking specified - providing mocked attendance data.")
            return

        if attendance_source["type"] == "benchmark":
            logger.info("Connecting to Benchmark to retrieve attendance data.")
            self._get_benchmark_attendance(attendance_source)
        else:
            raise KeyError(f"Student attendance of type: "
                           f"{attendance_source['type']} not implemented.")

    def _get_benchmark_attendance(self, attendance_source: dict):
        """Collect data from the UoM CS Benchmark API. To simplify the API requests, all
        of the attendance data for a particular course is downloaded at once. This means that
        StudentAttendance should be instantiated at the beginning and then data collected
        student by student by accessing the process_benchmark_data method.

        :param attendance_source: Information about where to get attendance data from.
        """
        headers = {'Private-Token': attendance_source["token"]}
        data = requests.get(attendance_source["url"], headers=headers).text
        self.data = json.loads(data)

    def get_student_attendance(self, student_id: str) -> int:
        """For an individual student, get their attendance from the list of all attendances.

        :param student_id: The university ID name of the student to get attendance of.
        :return student_attendance: The number of sessions attended in the current year.
        """
        if self.mock:
            return 8

        student_attendance = 0
        current_time = time.time()
        for week in self.data:
            # Only collect attendance data for weeks that have passed.
            if week["finish"] < current_time:
                try:
                    if week["events"][student_id][0]["data"] == "present":
                        student_attendance += 1
                except KeyError:
                    pass
        return student_attendance

    def _get_number_of_sessions(self) -> int:
        """Get the total number of sessions that a student could have attended in the
        current year."""

        if self.mock:
            return 10
        else:
            num_sessions = 0
            current_time = time.time()
            for week in self.data:
                if week["finish"] < current_time:
                    num_sessions += 1
            return num_sessions
